# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dispatch DataBase script."""

import copy
from datetime import datetime

import networkx as nx
import simplejson

import covalent.executor as covalent_executor
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.util_classes import Status

app_log = logger.app_log
log_stack_info = logger.log_stack_info
# DB Schema:
# TABLE dispatches
# * dispatch_id text primary key
# * result_dict text ; json-serialied dictionary representation of Result

# TODO: Move these to a common utils module


def extract_graph_node(node):
    # avoid mutating original node
    node = node.copy()

    # doc string
    f = node.get("function")
    if f is not None:
        node["doc"] = f.attrs["doc"]

    if "value" in node and node["value"] is not None:
        node["value"] = node["value"].object_string
    if "output" in node and node["output"] is not None:
        node["output"] = node["output"].object_string

    # metadata
    node["metadata"] = extract_metadata(node["metadata"])

    # prevent JSON encoding
    node["kwargs"] = encode_dict(node.get("kwargs"))

    # remove unused fields
    node.pop("function", None)
    node.pop("node_name", None)

    return node


def extract_metadata(metadata: dict):
    # TODO: This is an outdated method

    try:
        # avoid mutating original metadata
        metadata = copy.deepcopy(metadata)

        name = metadata["executor"]
        app_log.debug(f"Getting executor {name}")
        executor = covalent_executor._executor_manager.get_executor(name=name)

        if executor is not None:
            # extract attributes
            metadata["executor"] = encode_dict(executor.__dict__)
            if isinstance(name, str):
                metadata["executor_name"] = name
            else:
                metadata["executor_name"] = f"<{executor.__class__.__name__}>"

        metadata["hooks"]["deps"] = encode_dict(metadata["hooks"]["deps"])
        call_before = metadata["hooks"]["call_before"]
        call_after = metadata["hooks"]["call_after"]
        for i, dep in enumerate(call_before):
            call_before[i] = str(dep)

        for i, dep in enumerate(call_after):
            call_after[i] = str(dep)

        metadata["hooks"]["call_before"] = call_before
        metadata["hooks"]["call_after"] = call_after

    except (KeyError, AttributeError) as ex:
        app_log.debug(f"Exception when trying to extract metadata: {ex}")

    return metadata


def encode_dict(d):
    """Avoid JSON encoding when python str() suffices"""
    if not isinstance(d, dict):
        return d
    return {k: str(v) for (k, v) in d.items()}


def extract_graph(graph):
    graph = nx.json_graph.node_link_data(graph)
    nodes = list(map(extract_graph_node, graph["nodes"]))
    return {
        "nodes": nodes,
        "links": graph["links"],
    }


def result_encoder(obj):
    if isinstance(obj, Status):
        return obj.STATUS
    if isinstance(obj, datetime):
        return obj.isoformat()
    return str(obj)


def encode_result(result_obj):
    lattice = result_obj.lattice

    result_string = result_obj.encoded_result.json
    if not result_string:
        result_string = result_obj.encoded_result.object_string

    named_args = {k: v.object_string for k, v in lattice.named_args.items()}
    named_kwargs = {k: v.object_string for k, v in lattice.named_kwargs.items()}
    result_dict = {
        "dispatch_id": result_obj.dispatch_id,
        "status": result_obj.status,
        "result": result_string,
        "start_time": result_obj.start_time,
        "end_time": result_obj.end_time,
        "results_dir": result_obj.results_dir,
        "error": result_obj.error,
        "lattice": {
            "function_string": lattice.workflow_function_string,
            "doc": lattice.__doc__,
            "name": lattice.__name__,
            "inputs": encode_dict({**named_args, **named_kwargs}),
            "metadata": extract_metadata(lattice.metadata),
        },
        "graph": extract_graph(result_obj.lattice.transport_graph._graph),
    }

    jsonified_result = simplejson.dumps(result_dict, default=result_encoder, ignore_nan=True)

    return jsonified_result


class DispatchDB:
    """
    Wrapper for the database of workflows.
    """

    def __init__(self, dbpath: str = None) -> None:
        if dbpath:
            self._dbpath = dbpath
        else:
            self._dbpath = get_config("user_interface.dispatch_db")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return False



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines interactions with Qiskit Runtime sessions and services.
"""
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, Union

from qiskit_ibm_runtime import QiskitRuntimeService, Session


@lru_cache
def init_runtime_service(
    *,
    ibmqx_token: str = None,
    ibmqx_url: str = None,
    channel: str = "",
    instance: str = "",
    cloud_instance: str = "",
    hub: str = "",
    group: str = "",
    project: str = "",
) -> QiskitRuntimeService:
    """
    Start `QiskitRuntimeService` with specified settings.
    """

    if channel == "ibm_quantum":
        if not instance:
            # Combine hub, group, and project to produce the instance.
            hgp = [hub, group, project]
            if all(hgp):
                instance = "/".join(hgp)
            else:
                instance = "ibm-q/open/main"

    elif channel == "ibm_cloud":
        # Require `cloud_instance` to be specified in this case.
        if not cloud_instance:
            raise ValueError("The `cloud_instance` is required for the 'ibm_cloud' channel.")

        instance = cloud_instance

    else:
        raise ValueError(
            "Invalid `channel` argument, must be either 'ibm_quantum' or 'ibm_cloud'."
        )

    # Initialize the runtime service instance.
    return QiskitRuntimeService(
        channel=channel, token=ibmqx_token, url=ibmqx_url, instance=instance
    )


@dataclass(frozen=True)
class SessionIdentifier:
    """
    Proxy for defining a unique `Session` instance.
    """

    service_channel: str
    service_instance: str
    service_url: str
    backend_name: str
    max_time: Union[int, None]


def get_cached_session(service, backend, max_time) -> Session:
    """
    Global Qiskit IBM Runtime sessions, unique up to fields in `SessionIdentifier`
    """
    session_id = make_session_id(service, backend, max_time)
    if session_id not in _sessions:
        _sessions[session_id] = Session(
            service=service,
            backend=backend,
            max_time=max_time,
        )

    return _sessions[session_id]


def make_session_id(service, backend, max_time) -> SessionIdentifier:
    """
    Create session identifier from `Session` initialization arguments
    """
    return SessionIdentifier(
        # pylint: disable=protected-access
        service_channel=service._account.channel,
        service_instance=service._account.instance,
        service_url=service._account.url,
        backend_name=backend.name,
        max_time=max_time,
    )


_sessions: Dict[SessionIdentifier, Session] = {}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Covalent CLI Tool - Service Management."""


import asyncio
import contextlib
import json
import os
import shutil
import signal
import socket
import sys
import time
import traceback
import warnings
from pathlib import Path
from subprocess import DEVNULL, Popen
from typing import Optional

import click
import dask.system
import psutil
import requests
import sqlalchemy
from dask.distributed import Client
from distributed.comm import unparse_address
from distributed.comm.core import CommClosedError
from distributed.core import connect, rpc
from furl import furl
from natsort import natsorted
from rich.box import ROUNDED
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Prompt
from rich.status import Status
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from sqlalchemy import exc as sa_exc

from covalent._shared_files.config import ConfigManager, get_config, reload_config, set_config

from .._db.datastore import DataStore
from .migrate import migrate_pickled_result_object

UI_PIDFILE = get_config("dispatcher.cache_dir") + "/ui.pid"
UI_LOGFILE = get_config("user_interface.log_dir") + "/covalent_ui.log"
UI_SRVDIR = f"{os.path.dirname(os.path.abspath(__file__))}/../../covalent_ui"

MIGRATION_WARNING_MSG = "Covalent not started. The database needs to be upgraded."
MIGRATION_COMMAND_MSG = (
    '   (use "covalent db migrate" to run database migrations and then retry "covalent start")'
)
ZOMBIE_PROCESS_STATUS_MSG = "Covalent server is unhealthy: Process is in zombie status"
STOPPED_PROCESS_STATUS_MSG = "Covalent server is unhealthy: Process is in stopped status"

# Ignore SQLAlchemy warnings
warnings.simplefilter("ignore", category=sa_exc.SAWarning)


def print_header(console):
    branding_title = Text("Covalent", style="bold blue")
    github_link = Text("GitHub: https://github.com/AgnostiqHQ/covalent", style="cyan")
    docs_link = Text("Docs:   https://docs.covalent.xyz", style="cyan")
    console.print(Panel.fit(branding_title, padding=(1, 20)))
    console.print(github_link)
    console.print(docs_link)
    console.print()


def print_footer(console):
    console.print("\nFor additional help, use 'covalent --help'")


def _read_pid(filename: str) -> int:
    """
    Read the process ID from file.

    Args:
        filename: The path to the file to read the process ID from.

    Returns:
        The process ID.
    """

    try:
        pid = int(open(filename, "r").readline())
    except FileNotFoundError:
        pid = -1

    return pid


def _rm_pid_file(filename: str) -> None:
    """
    Remove a process ID file safely.

    Args:
        filename: The path to the file that will be removed.

    Returns:
        None
    """

    if os.path.isfile(filename):
        os.remove(filename)


def _port_from_pid(pid: int) -> Optional[int]:
    """
    Return the port in use by a process.

    Args:
        pid: Process ID.

    Returns:
        port: Port in use by the process.
    """

    if psutil.pid_exists(pid):
        return psutil.Process(pid).connections()[0].laddr.port
    return None


def _next_available_port(requested_port: int) -> int:
    """
    Return the next available port not in use.

    Args:
        requested_port: Port requested for a socket.

    Returns:
        assigned_port: Next available port greater than or equal to the requested port.
    """

    avail_port_found = False
    try_port = requested_port

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    while not avail_port_found:
        try:
            sock.bind(("localhost", try_port))
            avail_port_found = True
        except:
            try_port += 1

    sock.close()
    assigned_port = try_port

    if assigned_port != requested_port:
        click.echo(
            f"Port {requested_port} was already in use. Using port {assigned_port} instead."
        )

    return assigned_port


def _is_server_running() -> bool:
    """Check status of the Covalent server.

    Returns:
        status: Status of whether the server is running.
    """
    return _read_pid(UI_PIDFILE) != -1


def _graceful_start(
    server_root: str,
    pidfile: str,
    logfile: str,
    port: int,
    no_cluster: bool,
    develop: bool = False,
    no_triggers: bool = False,
    triggers_only: bool = False,
) -> int:
    """
    Gracefully start a Fast API app.

    Args:
        server_root: Directory where app.py is located.
        pidfile: Process ID file for the server.
        logfile: Log file for the server.
        port: Port requested to be used by the server.
        no_cluster: Dask cluster is not used.
        develop: Start the server in developer mode.
        no_triggers: Start the server without Triggers endpoints.
        triggers_only: Start the server with only Triggers endpoints.

    Returns:
        port: Port assigned to the server.
    """

    pid = _read_pid(pidfile)
    if psutil.pid_exists(pid):
        port = get_config("dispatcher.port")
        return port

    _rm_pid_file(pidfile)

    if no_triggers and triggers_only:
        raise ValueError(
            "Options '--no-triggers' and '--triggers-only' are mutually exclusive, please choose one at most."
        )

    no_triggers_flag = "--no-triggers" if no_triggers else ""
    triggers_only_flag = "--triggers-only" if triggers_only else ""

    pypath = f"PYTHONPATH={UI_SRVDIR}/../tests:$PYTHONPATH" if develop else ""
    dev_mode_flag = "--develop" if develop else ""
    no_cluster_flag = "--no-cluster" if no_cluster else ""

    port = _next_available_port(port)
    launch_str = f"{pypath} {sys.executable} app.py {dev_mode_flag} --port {port} {no_cluster_flag} {no_triggers_flag} {triggers_only_flag}>> {logfile} 2>&1"

    proc = Popen(launch_str, shell=True, stdout=DEVNULL, stderr=DEVNULL, cwd=server_root)
    pid = proc.pid

    with open(pidfile, "w") as PIDFILE:
        PIDFILE.write(str(pid))

    # Wait until the server actually starts listening on the port
    dispatcher_addr = f"http://localhost:{port}"
    up = False
    while not up:
        try:
            requests.get(dispatcher_addr, timeout=1)
            up = True
        except requests.exceptions.ConnectionError:
            time.sleep(1)

    # Since the dispatcher process might update the config file with the Dask cluster's state,
    # we need to sync those changes with the CLI's ConfigManager instance. Otherwise the next
    # call to `set_config()` from this module would obliterate the Dask cluster state.
    reload_config()

    Path(get_config("dispatcher.cache_dir")).mkdir(parents=True, exist_ok=True)
    Path(get_config("dispatcher.results_dir")).mkdir(parents=True, exist_ok=True)
    Path(get_config("dispatcher.log_dir")).mkdir(parents=True, exist_ok=True)
    Path(get_config("user_interface.log_dir")).mkdir(parents=True, exist_ok=True)

    return port


def _terminate_child_processes(pid: int) -> None:
    """For a given process, find all the child processes and terminate them.

    Args:
        pid: Process ID file for the main server process.

    Returns:
        None
    """

    # Uvicorn
    leader = psutil.Process(pid).children()[0]

    # Dask
    children = psutil.Process(leader.pid).children(recursive=True)

    with contextlib.suppress(psutil.NoSuchProcess):
        leader.send_signal(signal.SIGINT)

    for child_proc in children:
        with contextlib.suppress(psutil.NoSuchProcess):
            child_proc.kill()

    psutil.wait_procs(children)
    leader.wait()


def _graceful_shutdown(pidfile: str) -> None:
    """
    Gracefully shut down a server given a process ID.

    Args:
        pidfile: Process ID file for the server.

    Returns:
        None
    """
    console = Console()
    pid = _read_pid(pidfile)
    if psutil.pid_exists(pid):
        proc = psutil.Process(pid)
        _terminate_child_processes(pid)

        with contextlib.suppress(psutil.NoSuchProcess):
            proc.terminate()
            proc.wait()

    else:
        console.print("[yellow]Covalent server was not running.[/yellow]\n")

    _rm_pid_file(pidfile)


@click.command()
@click.option("-d", "--develop", is_flag=True, help="Start local server in developer mode")
@click.option(
    "-p",
    "--port",
    default=get_config("dispatcher.port"),
    show_default=True,
    help="Local server port number",
)
@click.option(
    "-m",
    "--mem-per-worker",
    required=False,
    is_flag=False,
    type=str,
    help="""Memory limit per worker in GB.
              Provide strings like 1gb/1GB or 0 for no limits""".replace(
        "\n", ""
    ),
)
@click.option(
    "-n",
    "--workers",
    required=False,
    is_flag=False,
    type=int,
    help="Number of Dask workers",
)
@click.option(
    "-t",
    "--threads-per-worker",
    required=False,
    is_flag=False,
    type=int,
    help="Number of threads per Dask worker",
)
@click.option(
    "--ignore-migrations",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Start server without database migrations",
)
@click.option(
    "--no-cluster",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Start server without Dask cluster",
)
@click.option(
    "--no-triggers",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Start server without a triggers server",
)
@click.option(
    "--triggers-only",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Start only the triggers server",
)
@click.option("--no-header", is_flag=True, default=False, hidden=True)
@click.option("--no-footer", is_flag=True, default=False, hidden=True)
@click.pass_context
def start(
    ctx: click.Context,
    port: int,
    develop: bool,
    no_cluster: bool,
    mem_per_worker: str,
    threads_per_worker: int,
    workers: int,
    ignore_migrations: bool,
    no_triggers: bool,
    triggers_only: bool,
    no_header: bool,
    no_footer: bool,
) -> None:
    """
    Start a local server
    """

    console = Console()

    if not no_header:
        print_header(console)

    # Display a header with a border
    console.print(Panel("Starting Local Server", expand=False, border_style="blue"))
    console.print()

    if os.environ.get("COVALENT_DEBUG_MODE") == "1":
        develop = True

    if os.environ.get("COVALENT_DISABLE_DASK") == "1":
        no_cluster = True

    if develop:
        set_config({"sdk.log_level": "debug"})

    db = DataStore.factory()

    # No migrations have run as of yet - run them automatically
    if not ignore_migrations and db.current_revision() is None:
        with Status("Running migrations...", console=console):
            db.run_migrations(logging_enabled=False)

    if db.is_migration_pending and not ignore_migrations:
        console.print(MIGRATION_WARNING_MSG, style="yellow")
        console.print(MIGRATION_COMMAND_MSG)
        console.print()
        return ctx.exit(1)

    if ignore_migrations and db.is_migration_pending:
        console.print(
            'Warning: Ignoring migrations is not recommended and may have unanticipated side effects. Use "covalent db migrate" to run migrations.',
            style="yellow",
        )
        console.print()

    set_config("user_interface.port", port)
    set_config("dispatcher.port", port)

    if not no_cluster:
        if mem_per_worker:
            set_config("dask.mem_per_worker", mem_per_worker)
        if threads_per_worker:
            set_config("dask.threads_per_worker", threads_per_worker)
        if workers:
            set_config("dask.num_workers", workers)

        set_config("sdk.no_cluster", "false")
    else:
        set_config("sdk.no_cluster", "true")

    try:
        with Status("Starting server...", console=console):
            port = _graceful_start(
                UI_SRVDIR,
                UI_PIDFILE,
                UI_LOGFILE,
                port,
                no_cluster,
                develop,
                no_triggers,
                triggers_only,
            )
    except Exception:
        click.secho("Error: ", fg="red")
        click.secho(
            "Covalent was unable to start due to the following error: ", fg="red", bold=True
        )
        click.secho(traceback.format_exc(), fg="lightgrey")

    set_config("user_interface.port", port)
    set_config("dispatcher.port", port)

    # Display server configuration in a table
    config_table = Table(title="Covalent Server Configuration", box=ROUNDED, show_header=False)
    config_table.add_column("Option", style="bold", no_wrap=True)
    config_table.add_column("Value")

    config_table.add_row(
        "Dispatcher Address", Text(str(get_config("dispatcher.address")), style="green")
    )
    config_table.add_row("Port", Text(str(port), style="green"))
    config_table.add_row("Develop", Text(str(develop), style="blue" if develop else "green"))
    config_table.add_row(
        "Disable Dask", Text(str(no_cluster), style="blue" if no_cluster else "green")
    )
    config_table.add_row("Memory per Worker", Text(str(mem_per_worker), style="magenta"))
    config_table.add_row("Threads per Worker", Text(str(threads_per_worker), style="magenta"))
    config_table.add_row("Workers", Text(str(workers), style="magenta"))
    config_table.add_row(
        "Ignore Migrations",
        Text(str(ignore_migrations), style="yellow" if ignore_migrations else "green"),
    )
    config_table.add_row(
        "Disable Triggers", Text(str(no_triggers), style="blue" if no_triggers else "green")
    )
    config_table.add_row(
        "Triggers Only", Text(str(triggers_only), style="blue" if triggers_only else "green")
    )

    console.print(config_table)
    console.print("\nServer Status: [green]:heavy_check_mark:[/green] Running", style="bold")

    dispatcher_address = f"http://{str(get_config('dispatcher.address'))}:{str(port)}"
    console.print(f"\nCovalent UI can be accessed at {dispatcher_address}")

    if not no_footer:
        console.print("\nFor a summary of the system status, use 'covalent status'")
        print_footer(console)


@click.command()
@click.option("--no-header", is_flag=True, default=False, hidden=True)
@click.option("--no-footer", is_flag=True, default=False, hidden=True)
def stop(no_header: bool, no_footer: bool) -> None:
    """
    Stop a local server
    """

    console = Console()
    if not no_header:
        print_header(console)

    console.print(Panel("Stopping Local Server", expand=False, border_style="blue"))
    console.print()

    with Status("Stopping server...", console=console):
        _graceful_shutdown(UI_PIDFILE)

    console.print("Server status: [red]:heavy_multiplication_x:[/red] Stopped", style="bold")

    if not no_footer:
        print_footer(console)


@click.command()
@click.option(
    "-p",
    "--port",
    default=None,
    type=int,
    help="Restart local server on given port",
)
@click.option("-d", "--develop", is_flag=True, help="Start local server in developer mode")
@click.option(
    "--no-cluster",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Restart server without Dask cluster",
)
@click.option(
    "--with-cluster",
    is_flag=True,
    required=False,
    show_default=True,
    default=False,
    help="Restart server with Dask cluster",
)
@click.pass_context
def restart(ctx, port: bool, develop: bool, no_cluster: bool, with_cluster: bool) -> None:
    """
    Restart a local server
    """

    if no_cluster and with_cluster:
        raise ValueError(
            "Options '--no-cluster' and '--with-cluster' are mutually exclusive, please choose one at most."
        )

    if no_cluster:
        set_config("sdk.no_cluster", "true")

    if with_cluster:
        set_config("sdk.no_cluster", "false")

    no_cluster_map = {"true": True, "false": False}
    configuration = {
        "port": port or get_config("dispatcher.port"),
        "develop": develop or (get_config("sdk.log_level") == "debug"),
        "no_cluster": no_cluster_map[get_config("sdk.no_cluster")],
        "mem_per_worker": get_config("dask.mem_per_worker"),
        "threads_per_worker": get_config("dask.threads_per_worker"),
        "workers": get_config("dask.num_workers"),
        "no_header": True,
    }

    ctx.invoke(stop, no_footer=True)
    console = Console()
    console.print()
    ctx.invoke(start, **configuration)


@click.command()
def status() -> None:
    """
    Display local server status
    """

    console = Console()
    print_header(console)

    console.print(Panel("Service Status", expand=False, border_style="blue"))
    console.print()

    with Status("Checking Covalent's Process ID...", console=console):
        pid = _read_pid(UI_PIDFILE)
        port = get_config("dispatcher.port")
        exists = psutil.pid_exists(pid)

    status_table = Table()
    status_table.add_column("Component", style="bold")
    status_table.add_column("Status", style="bold")

    if exists and pid != -1:
        status_table.add_row(
            "Covalent Server", f"[green]Running[/green] at http://localhost:{port}"
        )
    elif exists and psutil.Process(pid).status() == psutil.STATUS_ZOMBIE:
        status_table.add_row(
            "Covalent Server", "[yellow]Zombie process :zombie:[/yellow] - Recommend restart"
        )
    elif not exists or psutil.Process(pid).status() == psutil.STATUS_STOPPED:
        _rm_pid_file(UI_PIDFILE)
        status_table.add_row("Covalent Server", "[red]Stopped[/red]")

    if exists and pid != -1:
        if Path(get_config("dispatcher.heartbeat_file")).is_file():
            with open(get_config("dispatcher.heartbeat_file")) as f:
                last_seen = f.read().split(" ", 1)[1]
            status_table.add_row("", f"Last seen {last_seen}")
        response = requests.get(
            f"http://localhost:{port}/api/v1/dispatches/list?status_filter=RUNNING", timeout=1
        )
        running_workflows = response.json()["total_count"]
        status_table.add_row("", f"There are {running_workflows} workflows currently running.")

    admin_address = _get_cluster_admin_address()
    loop = asyncio.get_event_loop()
    cluster_status = (
        loop.run_until_complete(_get_cluster_status(admin_address)) if admin_address else None
    )

    if _is_server_running() and cluster_status:
        status_table.add_row("Dask Cluster", f"[green]Running[/green] at {admin_address}")
        client = Client(get_config("dask.scheduler_address"))
        running_tasks = len([task for k, v in client.processing().items() for task in v])
        status_table.add_row("", f"There are {running_tasks} tasks currently running.")
    else:
        status_table.add_row("Dask Cluster", "[red]Stopped[/red]")

    try:
        response = requests.get(f"http://localhost:{port}/api/triggers/status", timeout=1)
        trigger_status = response.json()["status"]
    except requests.exceptions.ConnectionError:
        trigger_status = "stopped"

    if trigger_status == "running":
        status_table.add_row("Triggers Server", "[green]Running[/green]")
    else:
        status_table.add_row("Triggers Server", "[red]Stopped[/red]")

    try:
        db = DataStore.factory()

        if db.is_migration_pending:
            status_table.add_row("Database", "[yellow]Migration pending[/yellow]")
        else:
            url = db.db_URL
            if os.environ.get("COVALENT_DATABASE_URL"):
                url = furl(url).origin
            status_table.add_row("Database", f"[green]Connected[/green] at {url}")

    except sqlalchemy.exc.OperationalError:
        status_table.add_row("Database", "[red]Disconnected[/red]")

    console.print(status_table, width=80)

    if cluster_status:
        console.print(
            "\nFor additional information about the Dask cluster, use 'covalent cluster --status'"
        )

    print_footer(console)


@click.command()
@click.option(
    "-H",
    "--hard",
    is_flag=True,
    help="Delete Covalent and workflow data. [default: False]",
)
@click.option(
    "-y", "--yes", is_flag=True, help="Approve without showing the warning. [default: False]"
)
@click.option("--hell-yeah", is_flag=True, hidden=True)
@click.pass_context
def purge(ctx, hard: bool, yes: bool, hell_yeah: bool) -> None:
    """
    Purge Covalent from this system
    """
    cm = ConfigManager()

    console = Console()

    removal_list = {
        get_config("sdk.log_dir"),
        get_config("dispatcher.cache_dir"),
        get_config("dispatcher.log_dir"),
        get_config("user_interface.log_dir"),
        os.path.dirname(cm.config_file),
    }

    if hell_yeah:
        hard = True
        yes = True

    if hard:
        removal_list.add(get_config("dispatcher.db_path"))

    if not yes:
        warning_text = Text("WARNING", style="bold yellow")
        warning_panel = Panel(warning_text, style="yellow", expand=False, padding=(0, 10))
        console.print(warning_panel)

        console.print("\nPurging will perform the following operations: ")

        console.print("0. Cancel all running workflows")
        console.print("1. Stop all services")

        for i, rem_path in enumerate(removal_list, start=2):
            if os.path.isdir(rem_path):
                console.print(f"{i}. {rem_path} directory will be deleted", style="red")
            else:
                console.print(f"{i}. {rem_path} file will be deleted", style="red")

        if hard:
            console.print("WARNING: All user data will be deleted", style="bold red")

        ans = Prompt.ask(  # Use Prompt.ask instead of console.Prompt.ask
            "\nAre you sure you want to continue?", choices=["y", "n"], default="n"
        )
        if ans == "n":
            console.print("Purge aborted.")
            print_footer(console)
            return

    # Shutdown covalent server
    console.print()
    ctx.invoke(stop, no_header=True, no_footer=True)

    # Remove all directories and files
    for rem_path in removal_list:
        if os.path.isdir(rem_path):
            shutil.rmtree(rem_path, ignore_errors=True)
        else:
            with contextlib.suppress(FileNotFoundError):
                os.remove(rem_path)

        console.print(f"Removed {rem_path}")

    print_footer(console)


@click.command()
def logs() -> None:
    """
    Display local server logs
    """
    console = Console()
    if os.path.exists(UI_LOGFILE):
        with open(UI_LOGFILE, "r") as logfile:
            log_content = logfile.read()
            syntax = Syntax(log_content, "log", theme="monokai", line_numbers=True, word_wrap=True)
            console.print(syntax)
    else:
        console.print(
            f"{UI_LOGFILE} not found. Restart the server to create a new log file.",
            style="bold red",
        )


@click.command()
@click.argument("result_pickle_path")
def migrate_legacy_result_object(result_pickle_path) -> None:
    """Migrate a legacy result object

    Example: `covalent migrate-legacy-result-object result.pkl`
    """

    migrate_pickled_result_object(result_pickle_path)


# Cluster CLI handlers (client side wrappers for the async handlers exposed
# in the dask cluster process)
async def _get_cluster_status(uri: str):
    """
    Returns status of all workers and scheduler in the cluster
    """

    try:
        async with rpc(uri, timeout=2) as r:
            cluster_status = await r.cluster_status()
    except (ConnectionRefusedError, CommClosedError, asyncio.exceptions.TimeoutError, OSError):
        return False
    return cluster_status


async def _get_cluster_address(uri):
    """
    Returns the TCP addresses of the scheduler and workers
    """
    async with rpc(uri, timeout=2) as r:
        addresses = await r.cluster_address()
    return addresses


async def _get_cluster_info(uri):
    """
    Return summary of cluster info
    """
    async with rpc(uri, timeout=2) as r:
        return await r.cluster_info()


async def _cluster_restart(uri):
    """
    Restart the cluster by individually restarting the cluster workers
    """
    async with rpc(uri, timeout=2) as r:
        await r.cluster_restart()


async def _cluster_scale(uri: str, nworkers: int):
    """
    Scale the cluster up/down depending on `nworkers`
    """
    comm = await connect(uri, timeout=2)
    await comm.write({"op": "cluster_scale", "size": nworkers})
    result = await comm.read()
    comm.close()
    return result


async def _get_cluster_size(uri) -> int:
    async with rpc(uri, timeout=2) as r:
        size = await r.cluster_size()
    return size


async def _get_cluster_logs(uri):
    """
    Retrieve the cluster logs from the scheduler directly
    """
    comm = await connect(uri, timeout=2)
    await comm.write({"op": "cluster_logs"})
    cluster_logs = await comm.read()
    comm.close()
    return cluster_logs


def _get_cluster_admin_address():
    try:
        admin_host = get_config("dask.admin_host")
        admin_port = get_config("dask.admin_port")
        admin_server_addr = unparse_address("tcp", f"{admin_host}:{admin_port}")
        return admin_server_addr
    except KeyError:
        return


@click.command()
@click.option("--status", is_flag=True, help="Display Dask cluster status")
@click.option("--info", is_flag=True, help="Display Dask cluster info")
@click.option(
    "--address", is_flag=True, help="Fetch connection information of the cluster scheduler/workers"
)
@click.option("--size", is_flag=True, help="Return number of active workers in the cluster")
@click.option("--restart", is_flag=True, help="Restart the cluster")
@click.option(
    "--scale",
    is_flag=False,
    nargs=1,
    type=int,
    default=dask.system.CPU_COUNT,
    show_default=True,
    help="Scale cluster by adding/removing workers to match `nworkers`",
)
@click.option("--logs", is_flag=True, default=False, help="Show Dask cluster logs")
def cluster(
    status: bool, info: bool, address: bool, size: bool, restart: bool, scale: int, logs: bool
):
    """
    Manage local server's Dask cluster
    """

    loop = asyncio.get_event_loop()
    admin_server_addr = _get_cluster_admin_address()

    console = Console()
    print_header(console)
    console.print(Panel("Covalent Dask Cluster", expand=False, border_style="blue"))
    console.print()

    def print_json(data):
        table = Table()
        table.add_column("Key")
        table.add_column("Value")

        if isinstance(data, dict):
            for key, value in data.items():
                table.add_row(key, str(value))
        elif isinstance(data, tuple):
            for idx, value in enumerate(data):
                table.add_row(str(idx), str(value))
        elif isinstance(data, int):
            table.add_row("Size", str(data))

        console.print(table)

    cluster_status = (
        loop.run_until_complete(_get_cluster_status(admin_server_addr))
        if admin_server_addr
        else None
    )
    if _is_server_running() and cluster_status:
        if status:
            print_json(dict(natsorted(cluster_status.items())))
            degraded = {k: v for k, v in cluster_status.items() if v != "running"}
            if degraded:
                console.print(
                    "\nDask Cluster Status: :face_with_thermometer: Running - Degraded",
                    style="bold",
                )
            else:
                console.print(
                    "\nDask Cluster Status: [green]:heavy_check_mark:[/green] Running - Healthy",
                    style="bold",
                )
            diagnostics_dashboard_addr = get_config("dask.dashboard_link")
            console.print(f"Diagnostics Dashboard: {diagnostics_dashboard_addr} (requires bokeh)")

        elif info:
            print_json(loop.run_until_complete(_get_cluster_info(admin_server_addr)))
        elif address:
            print_json(loop.run_until_complete(_get_cluster_address(admin_server_addr)))
        elif size:
            print_json(loop.run_until_complete(_get_cluster_size(admin_server_addr)))
        elif restart:
            with Status("Restarting the cluster...", spinner="dots") as status:
                loop.run_until_complete(_cluster_restart(admin_server_addr))
                status.update("Cluster restarted")
                console.print("\n")
                console.print(
                    Panel("Cluster restarted", box=ROUNDED, expand=False, border_style="green")
                )
        elif logs:
            console.print(
                "\n".join(
                    [
                        " ".join(x)
                        for x in loop.run_until_complete(_get_cluster_logs(admin_server_addr))[0]
                    ]
                )
            )
        elif scale:
            with Status("Scaling the cluster...", spinner="dots") as status:
                loop.run_until_complete(_cluster_scale(admin_server_addr, nworkers=scale))
                status.update(f"Cluster scaled to {scale} workers")
                console.print(
                    Panel(
                        f"Cluster scaled to {scale} workers",
                        box=ROUNDED,
                        expand=False,
                        border_style="green",
                    )
                )
    else:
        console.print("Dask Cluster Status: [red]:heavy_multiplication_x:[/red] Stopped")

    print_footer(console)


@click.command()
def config() -> None:
    """Display the Covalent configuration"""

    cm = ConfigManager()

    console = Console()
    print_header(console)
    console.print(Panel("Covalent Configuration", expand=False, border_style="blue"))
    console.print()

    cm.read_config()
    config_data = json.loads(json.dumps(cm.config_data, sort_keys=True))
    sorted_sections = ["sdk", "dispatcher", "dask", "executors"]

    for section in sorted_sections:
        keys = config_data[section]

        # Create a table for each section
        section_table = Table(title=f"[bold]{section}[/bold]", title_style="bold")
        section_table.add_column("Key", style="bold")
        section_table.add_column("Value", style="bold")

        for key in keys:
            section_table.add_row(key, str(config_data[section][key]))

        # Wrap the table in a panel
        section_panel = Panel(
            section_table, expand=False, border_style="blue", padding=(0, 1), width=80
        )

        console.print(section_panel)

    print_footer(console)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class corresponding to computation nodes."""

import inspect
import json
import operator
import tempfile
from builtins import list
from copy import deepcopy
from dataclasses import asdict
from functools import wraps
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Union

from covalent._dispatcher_plugins.local import LocalDispatcher

from .._file_transfer.enums import Order
from .._file_transfer.file_transfer import FileTransfer
from .._shared_files import logger
from .._shared_files.context_managers import active_lattice_manager
from .._shared_files.defaults import (
    WAIT_EDGE_NAME,
    DefaultMetadataValues,
    electron_dict_prefix,
    electron_list_prefix,
    parameter_prefix,
    prefix_separator,
    sublattice_prefix,
)
from .._shared_files.util_classes import RESULT_STATUS
from .._shared_files.utils import (
    filter_null_metadata,
    get_named_params,
    get_serialized_function_str,
)
from .depsbash import DepsBash
from .depscall import RESERVED_RETVAL_KEY__FILES, DepsCall
from .depsmodule import DepsModule
from .depspip import DepsPip
from .lattice import Lattice
from .transport import TransportableObject, encode_metadata

consumable_constraints = ["budget", "time_limit"]

DEFAULT_METADATA_VALUES = asdict(DefaultMetadataValues())

if TYPE_CHECKING:
    from ..executor import BaseExecutor
    from .transport import _TransportGraph

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class Electron:
    """
    An electron (or task) object that is a modular component of a
    work flow and is returned by :obj:`electron <covalent.electron>`.

    Attributes:
        function: Function to be executed.
        node_id: Node id of the electron.
        metadata: Metadata to be used for the function execution.
        kwargs: Keyword arguments if any.
        task_group_id: the group to which the task be assigned when it is bound to a graph node. If unset, the group id will default to node id.
        packing_tasks: Flag to indicate whether task packing is enabled.
    """

    def __init__(
        self,
        function: Callable,
        node_id: int = None,
        metadata: dict = None,
        task_group_id: int = None,
        packing_tasks: bool = False,
    ) -> None:
        if metadata is None:
            metadata = {}
        self.function = function
        self.node_id = node_id
        self.metadata = metadata
        self.task_group_id = task_group_id
        self._packing_tasks = packing_tasks
        self._function_string = get_serialized_function_str(function)

    @property
    def packing_tasks(self) -> bool:
        return self._packing_tasks

    def set_metadata(self, name: str, value: Any) -> None:
        """
        Function to add/edit metadata of given name and value
        to electron's metadata.

        Args:
            name: Name of the metadata to be added/edited.
            value: Value of the metadata to be added/edited.

        Returns:
            None
        """

        self.metadata[name] = value

    def get_metadata(self, name: str) -> Any:
        """
        Get value of the metadata of given name.

        Args:
            name: Name of the metadata whose value is needed.

        Returns:
            value: Value of the metadata of given name.

        Raises:
            KeyError: If metadata of given name is not present.
        """

        return self.metadata[name]

    def get_op_function(
        self, operand_1: Union[Any, "Electron"], operand_2: Union[Any, "Electron"], op: str
    ) -> "Electron":
        """
        Function to handle binary operations with electrons as operands.
        This will not execute the operation but rather create another electron
        which will be postponed to be executed according to the default electron
        configuration/metadata.

        This also makes sure that if these operations are being performed outside
        of a lattice, then they are performed as is.

        Args:
            operand_1: First operand of the binary operation.
            operand_2: Second operand of the binary operation.
            op: Operator to be used in the binary operation.

        Returns:
            electron: Electron object corresponding to the operation execution.
                      Behaves as a normal function call if outside a lattice.
        """
        op_table = {
            "+": operator.add,
            "-": operator.sub,
            "*": operator.mul,
            "/": operator.truediv,
            "**": operator.pow,
        }

        def rename(op1: Any, op: str, op2: Any) -> Callable:
            """
            Decorator to rename a function according
            to the operation being performed.

            Args:
                op1: First operand
                op: Operator
                op2: Second operand

            Returns:
                function: Renamed decorated function.
            """

            def decorator(f):
                op1_name = op1
                if hasattr(op1, "function") and op1.function:
                    op1_name = op1.function.__name__
                op2_name = op2
                if hasattr(op2, "function") and op2.function:
                    op2_name = op2.function.__name__

                f.__name__ = f"{op1_name}_{op}_{op2_name}"
                return f

            return decorator

        @rename(operand_1, op, operand_2)
        def func_for_op(arg_1: Union[Any, "Electron"], arg_2: Union[Any, "Electron"]) -> Any:
            """
            Intermediate function for the binary operation.

            Args:
                arg_1: First operand
                arg_2: Second operand

            Returns:
                result: Result of the binary operation.
            """

            return op_table[op](arg_1, arg_2)

        # Mint an arithmetic electron and execute it using the
        # enclosing lattice's workflow_executor.

        metadata = encode_metadata(DEFAULT_METADATA_VALUES.copy())
        executor = metadata["workflow_executor"]
        executor_data = metadata["workflow_executor_data"]

        op_electron = Electron(func_for_op, metadata=metadata)

        if active_lattice := active_lattice_manager.get_active_lattice():
            executor = active_lattice.metadata.get(
                "workflow_executor", metadata["workflow_executor"]
            )
            executor_data = active_lattice.metadata.get(
                "workflow_executor_data", metadata["workflow_executor_data"]
            )
            op_electron.metadata["executor"] = executor
            op_electron.metadata["executor_data"] = executor_data

        return op_electron(arg_1=operand_1, arg_2=operand_2)

    def __add__(self, other):
        return self.get_op_function(self, other, "+")

    def __radd__(self, other):
        return self.get_op_function(other, self, "+")

    def __sub__(self, other):
        return self.get_op_function(self, other, "-")

    def __rsub__(self, other):
        return self.get_op_function(other, self, "-")

    def __mul__(self, other):
        return self.get_op_function(self, other, "*")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.get_op_function(self, other, "/")

    def __rtruediv__(self, other):
        return self.get_op_function(other, self, "/")

    def __pow__(self, other):
        return self.get_op_function(self, other, "**")

    def __int__(self):
        return int()

    def __float__(self):
        return float()

    def __complex__(self):
        return complex()

    def _get_collection_electron(self, name: str, func: Callable, metadata: Dict) -> "Electron":
        """Get collection electron with task packing enabled.

        Args:
            name: Name of the collection node.
            func: Function to be executed.

        Returns:
            Electron object with task packing enabled.

        """

        active_lattice = active_lattice_manager.get_active_lattice()
        return (
            Electron(function=func, metadata=self.metadata.copy())
            if name.startswith(sublattice_prefix)
            else Electron(
                function=func,
                metadata=metadata,
                task_group_id=self.task_group_id,
                packing_tasks=True and active_lattice.task_packing,
            )
        )

    def __iter__(self):
        last_frame = inspect.currentframe().f_back
        bytecode = last_frame.f_code.co_code
        expected_unpack_values = bytecode[last_frame.f_lasti + 1]

        if expected_unpack_values < 2:
            return

        for i in range(expected_unpack_values):
            if active_lattice := active_lattice_manager.get_active_lattice():
                try:
                    node_name = prefix_separator + self.function.__name__ + "()" + f"[{i}]"

                except AttributeError:
                    # The case when nested iter calls are made on the same electron
                    node_name = prefix_separator + active_lattice.transport_graph.get_node_value(
                        self.node_id, "name"
                    )
                    node_name += f"[{i}]"

                def get_item(e, key):
                    return e[key]

                get_item.__name__ = node_name

                # Perform a deep copy so as to not modify the parent
                # electron's hooks
                iterable_metadata = deepcopy(self.metadata)

                filtered_call_before = []
                if "call_before" in iterable_metadata["hooks"]:
                    for elem in iterable_metadata["hooks"]["call_before"]:
                        if elem["attributes"]["retval_keyword"] != "files":
                            filtered_call_before.append(elem)
                    iterable_metadata["hooks"]["call_before"] = filtered_call_before

                # Pack with main electron unless it is a sublattice.
                name = active_lattice.transport_graph.get_node_value(self.node_id, "name")
                yield self._get_collection_electron(name, get_item, iterable_metadata)(self, i)

    def __getattr__(self, attr: str) -> "Electron":
        # This is to handle the cases where magic functions are attempted
        # to be accessed. For example, in the case of pickling, sometimes
        # __getstate__ is called and we don't want to return an electron
        # object in that case.
        if attr.startswith("__") and attr.endswith("__"):
            return super().__getattr__(attr)

        if attr == "keys":
            raise AttributeError(
                "`keys` attribute should not be used in Electron objects due to conflict with `dict.keys`",
                "Please change the name of the attribute you want to use.",
            )

        if active_lattice := active_lattice_manager.get_active_lattice():

            def get_attr(e, attr):
                return getattr(e, attr)

            get_attr.__name__ = prefix_separator + self.function.__name__ + ".__getattr__"

            # Pack with main electron except for sublattices
            name = active_lattice.transport_graph.get_node_value(self.node_id, "name")
            metadata = self.metadata.copy()
            bound_electron = self._get_collection_electron(name, get_attr, metadata)(self, attr)
            return bound_electron

        return super().__getattr__(attr)

    def __getitem__(self, key: Union[int, str]) -> "Electron":
        if active_lattice := active_lattice_manager.get_active_lattice():

            def get_item(e, key):
                return e[key]

            get_item.__name__ = prefix_separator + self.function.__name__ + ".__getitem__"
            name = active_lattice.transport_graph.get_node_value(self.node_id, "name")
            metadata = self.metadata.copy()
            return self._get_collection_electron(name, get_item, metadata)(self, key)

        raise StopIteration

    def __call__(self, *args, **kwargs) -> Union[Any, "Electron"]:
        """
        Function to execute the electron.

        This behaves differently if the execution call is made inside a lattice
        and just adds the electron as a node to the lattice's transport graph.

        If the execution call is made outside of a lattice, then it executes the
        electron as a normal function call.

        Also contains a postprocessing part where the lattice's function is executed
        after all the nodes in the lattice's transport graph are executed. Then the
        execution call to the electron is replaced by its corresponding result.

        Note: Bound electrons are defined as electrons with a valid node_id, since it means they are bound to a TransportGraph.
        """

        # Check if inside a lattice and if not, perform a direct invocation of the function
        active_lattice = active_lattice_manager.get_active_lattice()
        if active_lattice is None:
            return self.function(*args, **kwargs)

        if active_lattice.post_processing:
            output = active_lattice.electron_outputs[0]
            active_lattice.electron_outputs.pop(0)
            return output

        # Setting metadata for default values according to lattice's metadata.
        for k in self.metadata:
            if (
                k not in consumable_constraints
                and k in DEFAULT_METADATA_VALUES
                and self.get_metadata(k) is None
            ):
                meta = active_lattice.get_metadata(k) or DEFAULT_METADATA_VALUES[k]
                self.set_metadata(k, meta)

        # Handle replace_electrons for redispatch
        name = self.function.__name__
        if name in active_lattice.replace_electrons:
            # Temporarily pop the replacement to avoid infinite
            # recursion.
            replacement_electron = active_lattice.replace_electrons.pop(name)

            # TODO: check that replacement has the same
            # signature. Also, although electron -> sublattice or
            # sublattice -> electron are technically possible, these
            # replacements will not work with the "exhaustive"
            # postprocess method which requires that the number of nodes be
            # determined by the lattice inputs.

            # This will return a bound replacement electron
            bound_electron = replacement_electron(*args, **kwargs)
            active_lattice.transport_graph.set_node_value(
                bound_electron.node_id,
                "status",
                RESULT_STATUS.PENDING_REPLACEMENT,
            )

            active_lattice.replace_electrons[name] = replacement_electron
            return bound_electron

        # Handle sublattices by injecting _build_sublattice_graph node
        if isinstance(self.function, Lattice):
            parent_metadata = active_lattice.metadata.copy()
            app_log.debug(f"Parent lattice metadata: {parent_metadata}")
            e_meta = parent_metadata.copy()
            e_meta["executor"] = e_meta.pop("workflow_executor")
            e_meta["executor_data"] = e_meta.pop("workflow_executor_data")

            sub_electron = Electron(
                function=_build_sublattice_graph,
                metadata=e_meta,
            )

            name = sublattice_prefix + self.function.__name__
            function_string = self._function_string
            bound_electron = sub_electron(
                self.function, json.dumps(parent_metadata), *args, **kwargs
            )

            active_lattice.transport_graph.set_node_value(bound_electron.node_id, "name", name)
            active_lattice.transport_graph.set_node_value(
                bound_electron.node_id,
                "function_string",
                function_string,
            )

            return bound_electron

        # Add a node to the transport graph of the active lattice. Electrons bound to nodes will never be packed with the
        # 'master' Electron. # Add non-sublattice node to the transport graph of the active lattice.

        self.node_id = active_lattice.transport_graph.add_node(
            name=self.function.__name__,
            function=self.function,
            metadata=self.metadata.copy(),
            function_string=self._function_string,
            task_group_id=self.task_group_id if self.packing_tasks else None,
        )
        self.task_group_id = self.task_group_id if self.packing_tasks else self.node_id

        if self.function:
            named_args, named_kwargs = get_named_params(self.function, args, kwargs)

            # For positional arguments
            # We use the fact that as of Python 3.6, dict order == insertion order
            for arg_index, item in enumerate(named_args.items()):
                key, value = item
                self.connect_node_with_others(
                    self.node_id, key, value, "arg", arg_index, active_lattice.transport_graph
                )

            # For keyword arguments
            # Filter out kwargs to be injected by call_before call_deps during execution.

            retval_keywords = {}
            if "call_before" in self.metadata["hooks"]:
                call_before = self.metadata["hooks"]["call_before"]
                retval_keywords = {
                    item["attributes"]["retval_keyword"]: None for item in call_before
                }

            for key, value in named_kwargs.items():
                if key in retval_keywords:
                    app_log.debug(
                        f"kwarg {key} for function {self.function.__name__} to be injected at runtime"
                    )
                    continue
                self.connect_node_with_others(
                    self.node_id, key, value, "kwarg", None, active_lattice.transport_graph
                )

        bound_electron = Electron(
            self.function,
            metadata=self.metadata,
            node_id=self.node_id,
            task_group_id=self.task_group_id,
            packing_tasks=self.packing_tasks,
        )
        active_lattice._bound_electrons[self.node_id] = bound_electron
        return bound_electron

    def connect_node_with_others(
        self,
        node_id: int,
        param_name: str,
        param_value: Union[Any, "Electron"],
        param_type: str,
        arg_index: int,
        transport_graph: "_TransportGraph",
    ):
        """
        Adds a node along with connecting edges for all the arguments to the electron.

        Args:
            node_id: Node number of the electron
            param_name: Name of the parameter
            param_value: Value of the parameter
            param_type: Type of parameter, positional or keyword
            transport_graph: Transport graph of the lattice

        Returns:
            None
        """

        collection_metadata = encode_metadata(DEFAULT_METADATA_VALUES.copy())
        active_lattice = active_lattice_manager.get_active_lattice()

        if "executor" in self.metadata:
            collection_metadata["executor"] = self.metadata["executor"]
            collection_metadata["executor_data"] = self.metadata["executor_data"]

        if isinstance(param_value, Electron):
            transport_graph.add_edge(
                param_value.node_id,
                node_id,
                edge_name=param_name,
                param_type=param_type,
                arg_index=arg_index,
            )

        elif isinstance(param_value, (list, tuple, set)):
            # Tuples and sets will also be converted to lists
            def _auto_list_node(*args, **kwargs):
                return list(args)

            list_electron = Electron(
                function=_auto_list_node,
                metadata=collection_metadata,
                task_group_id=self.task_group_id,
                packing_tasks=True and active_lattice.task_packing,
            )  # Group the auto-generated node with the main node.
            bound_electron = list_electron(*param_value)
            transport_graph.set_node_value(bound_electron.node_id, "name", electron_list_prefix)
            transport_graph.add_edge(
                list_electron.node_id,
                node_id,
                edge_name=param_name,
                param_type=param_type,
                arg_index=arg_index,
            )

        elif isinstance(param_value, dict):

            def _auto_dict_node(keys, values):
                return {keys[i]: values[i] for i in range(len(keys))}

            dict_electron = Electron(
                function=_auto_dict_node,
                metadata=collection_metadata,
                task_group_id=self.task_group_id,
                packing_tasks=True and active_lattice.task_packing,
            )  # Group the auto-generated node with the main node.
            bound_electron = dict_electron(list(param_value.keys()), list(param_value.values()))
            transport_graph.set_node_value(bound_electron.node_id, "name", electron_dict_prefix)
            transport_graph.add_edge(
                dict_electron.node_id,
                node_id,
                edge_name=param_name,
                param_type=param_type,
                arg_index=arg_index,
            )

        else:
            encoded_param_value = TransportableObject.make_transportable(param_value)

            parameter_node = transport_graph.add_node(
                name=parameter_prefix + str(param_value),
                function=None,
                metadata=encode_metadata(DEFAULT_METADATA_VALUES.copy()),
                value=encoded_param_value,
                output=encoded_param_value,
            )
            transport_graph.add_edge(
                parameter_node,
                node_id,
                edge_name=param_name,
                param_type=param_type,
                arg_index=arg_index,
            )

        return node_id

    def wait_for(self, electrons: Union["Electron", Iterable["Electron"]]):
        """
        Waits for the given electrons to complete before executing this one.
        Adds the necessary edges between this and those electrons without explicitly
        connecting their inputs/outputs.

        Useful when execution of this electron relies on a side-effect from the another one.

        Args:
            electrons: Electron(s) which will be waited for to complete execution
                       before starting execution for this one

        Returns:
            Electron
        """

        active_lattice = active_lattice_manager.get_active_lattice()

        # Just using list(electrons) will not work since we are overriding the __iter__
        # method for an Electron which results in it essentially disappearing, thus using
        # [electrons] to create the list if there's a single electron
        electrons = [electrons] if isinstance(electrons, Electron) else list(electrons)

        for el in electrons:
            active_lattice.transport_graph.add_edge(
                el.node_id,
                self.node_id,
                edge_name=WAIT_EDGE_NAME,
            )

        return Electron(
            self.function,
            metadata=self.metadata,
            node_id=self.node_id,
        )

    @property
    def as_transportable_dict(self) -> Dict:
        """Get transportable electron object and metadata."""
        return {
            "name": self.function.__name__,
            "function": TransportableObject(self.function).to_dict(),
            "function_string": get_serialized_function_str(self.function),
            "metadata": filter_null_metadata(self.metadata),
        }


# Dynamically adding properties to the Electron class
# This is being done this way so that it's easier to add
# or remove properties in the future without having to
# edit the class definition itself.
Electron.executor = property(lambda self: self.get_metadata("executor"))
Electron.executor = Electron.executor.setter(
    lambda self, value: self.metadata.update(encode_metadata({"executor": value}))
)


def electron(
    _func: Optional[Callable] = None,
    *,
    backend: Optional[str] = None,
    executor: Optional[Union[List[Union[str, "BaseExecutor"]], Union[str, "BaseExecutor"]]] = None,
    # Add custom metadata fields here
    files: List[FileTransfer] = [],
    deps_bash: Union[DepsBash, List, str] = None,
    deps_pip: Union[DepsPip, list] = None,
    deps_module: Union[DepsModule, List[DepsModule], str, List[str]] = None,
    call_before: Union[List[DepsCall], DepsCall] = None,
    call_after: Union[List[DepsCall], DepsCall] = None,
) -> Callable:  # sourcery skip: assign-if-exp
    """
    Electron decorator to be called upon a function. Returns the wrapper function with the same functionality as `_func`.

    Args:
        _func: function to be decorated

    Keyword Args:
        backend: DEPRECATED: Same as `executor`.
        executor: Alternative executor object to be used by the electron execution. If not passed, the dask
            executor is used by default.
        deps_bash: An optional DepsBash object specifying a list of shell commands to run before `_func`
        deps_pip: An optional DepsPip object specifying a list of PyPI packages to install before running `_func`
        deps_module: An optional DepsModule (or similar) object specifying which user modules to load before running `_func`
        call_before: An optional list of DepsCall objects specifying python functions to invoke before the electron
        call_after: An optional list of DepsCall objects specifying python functions to invoke after the electron
        files: An optional list of FileTransfer objects which copy files to/from remote or local filesystems.

    Returns:
        :obj:`Electron <covalent._workflow.electron.Electron>` : Electron object inside which the decorated function exists.

    """

    if backend:
        app_log.warning(
            "backend is deprecated and will be removed in a future release. Please use executor keyword instead.",
            exc_info=DeprecationWarning,
        )
        executor = backend

    deps = {} if deps_bash or deps_pip else None

    if isinstance(deps_bash, DepsBash):
        deps["bash"] = deps_bash
    if isinstance(deps_bash, (list, str)):
        deps["bash"] = DepsBash(commands=deps_bash)

    internal_call_before_deps = []
    internal_call_after_deps = []

    if files:
        for file_transfer in files:
            _file_transfer_pre_hook_, _file_transfer_call_dep_ = file_transfer.cp()

            # pre-file transfer hook to create any necessary temporary files
            internal_call_before_deps.append(
                DepsCall(
                    _file_transfer_pre_hook_,
                    retval_keyword=RESERVED_RETVAL_KEY__FILES,
                    override_reserved_retval_keys=True,
                )
            )

            if file_transfer.order == Order.AFTER:
                internal_call_after_deps.append(DepsCall(_file_transfer_call_dep_))
            else:
                internal_call_before_deps.append(DepsCall(_file_transfer_call_dep_))

    if deps_module:
        if isinstance(deps_module, list):
            # Convert to DepsModule objects
            converted_deps = []
            for dep in deps_module:
                if type(dep) in [str, ModuleType]:
                    converted_deps.append(DepsModule(dep))
                else:
                    converted_deps.append(dep)
            deps_module = converted_deps

        elif type(deps_module) in [str, ModuleType]:
            deps_module = [DepsModule(deps_module)]

        elif isinstance(deps_module, DepsModule):
            deps_module = [deps_module]

        internal_call_before_deps.extend(deps_module)

    if isinstance(deps_pip, DepsPip):
        deps["pip"] = deps_pip
    if isinstance(deps_pip, list):
        deps["pip"] = DepsPip(packages=deps_pip)

    if isinstance(call_before, DepsCall):
        call_before = [call_before]

    if isinstance(call_after, DepsCall):
        call_after = [call_after]

    call_before_final = [] if internal_call_before_deps or call_before else None
    if internal_call_before_deps:
        call_before_final.extend(internal_call_before_deps)
    if call_before:
        call_before_final.extend(call_before)

    call_after_final = [] if internal_call_after_deps or call_after else None
    if internal_call_after_deps:
        call_after_final.extend(internal_call_after_deps)
    if call_after:
        call_after_final.extend(call_after)

    if deps is None and call_before_final is None and call_after_final is None:
        hooks = None
    else:
        hooks = {}
        if deps is not None:
            hooks["deps"] = deps
        if call_before_final is not None:
            hooks["call_before"] = call_before_final
        if call_after_final is not None:
            hooks["call_after"] = call_after_final

    constraints = {
        "executor": executor,
        "hooks": hooks,
    }
    constraints = encode_metadata(constraints)

    def decorator_electron(func=None):
        """
        Electron decorator function. Note that the electron_object
        defined below is an example of an unbound electron, i.e.
        electron without a node id.

        """

        electron_object = Electron(func)
        for k, v in constraints.items():
            electron_object.set_metadata(k, v)
        electron_object.__doc__ = func.__doc__

        @wraps(func)
        def wrapper(*args, **kwargs):
            return electron_object(*args, **kwargs)

        wrapper.electron_object = electron_object

        return wrapper

    if _func is None:  # decorator is called with arguments
        return decorator_electron
    else:  # decorator is called without arguments
        return decorator_electron(_func)


def wait(child, parents):
    """Instructs Covalent that an electron should wait for some other
    tasks to complete before it is dispatched.

    Args:
        child: the dependent electron
        parents: Electron(s) which must complete before `waiting_electron` starts

    Returns:
        waiting_electron

    Useful when execution of an electron relies on a side-effect
    from another one.

    """
    active_lattice = active_lattice_manager.get_active_lattice()

    if active_lattice and not active_lattice.post_processing:
        return child.wait_for(parents)
    else:
        return child


@electron
def to_decoded_electron_collection(**x):
    """Interchanges order of serialize -> collection"""
    collection = list(x.values())[0]
    if isinstance(collection, list):
        return TransportableObject.deserialize_list(collection)
    elif isinstance(collection, dict):
        return TransportableObject.deserialize_dict(collection)


# Copied from runner.py
def _build_sublattice_graph(sub: Lattice, json_parent_metadata: str, *args, **kwargs):
    import os

    parent_metadata = json.loads(json_parent_metadata)
    for k in sub.metadata.keys():
        if not sub.metadata[k] and k != "triggers":
            sub.metadata[k] = parent_metadata[k]

    sub.build_graph(*args, **kwargs)

    try:
        # Attempt multistage sublattice dispatch. For now we require
        # the executor to reach the Covalent server
        parent_dispatch_id = os.environ["COVALENT_DISPATCH_ID"]
        dispatcher_url = os.environ["COVALENT_DISPATCHER_URL"]

        with tempfile.TemporaryDirectory(prefix="covalent-") as staging_path:
            manifest = LocalDispatcher.prepare_manifest(sub, staging_path)

            # Omit these two steps to return the manifest to Covalent and
            # request the assets be pulled
            recv_manifest = LocalDispatcher.register_manifest(
                manifest,
                dispatcher_addr=dispatcher_url,
                parent_dispatch_id=parent_dispatch_id,
                push_assets=True,
            )
            LocalDispatcher.upload_assets(recv_manifest)

        return recv_manifest.model_dump_json()

    except Exception as ex:
        # Fall back to legacy sublattice handling
        print("Falling back to legacy sublattice handling")
        return sub.serialize_to_json()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class corresponding to computation workflow."""

import importlib.metadata
import json
import warnings
import webbrowser
from builtins import list
from copy import deepcopy
from dataclasses import asdict
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from .._shared_files import logger
from .._shared_files.config import get_config
from .._shared_files.context_managers import active_lattice_manager
from .._shared_files.defaults import DefaultMetadataValues
from .._shared_files.utils import get_named_params, get_serialized_function_str, get_ui_url
from .depsbash import DepsBash
from .depscall import DepsCall
from .depspip import DepsPip
from .postprocessing import Postprocessor
from .transport import (
    TransportableObject,
    _TransportGraph,
    add_module_deps_to_lattice_metadata,
    encode_metadata,
)

if TYPE_CHECKING:
    from .._results_manager.result import Result
    from ..executor import BaseExecutor
    from ..triggers import BaseTrigger

from .._shared_files.utils import get_imports, get_serialized_function_str

consumable_constraints = []

DEFAULT_METADATA_VALUES = asdict(DefaultMetadataValues())

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class Lattice:
    """
    A lattice workflow object that holds the work flow graph and is returned by :obj:`lattice <covalent.lattice>` decorator.

    Attributes:
        workflow_function: The workflow function that is decorated by :obj:`lattice <covalent.lattice>` decorator.
        transport_graph: The transport graph which will be the basis on how the workflow is executed.
        metadata: Dictionary of metadata of the lattice.
        post_processing: Boolean to indicate if the lattice is in post processing mode or not.
        kwargs: Keyword arguments passed to the workflow function.
        electron_outputs: Dictionary of electron outputs received after workflow execution.
    """

    def __init__(
        self, workflow_function: Callable, transport_graph: _TransportGraph = None
    ) -> None:
        self.workflow_function = workflow_function
        self.workflow_function_string = get_serialized_function_str(self.workflow_function)
        self.transport_graph = transport_graph or _TransportGraph()
        self.metadata = {}
        self.__name__ = self.workflow_function.__name__
        self.__doc__ = self.workflow_function.__doc__
        self.post_processing = False
        self.inputs = None
        self.named_args = None
        self.named_kwargs = None
        self.electron_outputs = {}
        self.lattice_imports, self.cova_imports = get_imports(self.workflow_function)

        self.workflow_function = TransportableObject.make_transportable(self.workflow_function)

        # Bound electrons are defined as electrons with a valid node_id, since it means they are bound to a TransportGraph.
        self._bound_electrons = {}  # Clear before serializing

        self.python_version = self.workflow_function.python_version
        self.covalent_version = importlib.metadata.version("covalent")

    # To be called after build_graph
    def serialize_to_json(self) -> str:
        attributes = deepcopy(self.__dict__)
        attributes["workflow_function"] = self.workflow_function.to_dict()

        attributes["metadata"] = encode_metadata(self.metadata)
        attributes["transport_graph"] = None
        if self.transport_graph:
            attributes["transport_graph"] = self.transport_graph.serialize_to_json()

        attributes["inputs"] = self.inputs.to_dict()
        attributes["named_args"] = self.named_args.to_dict()
        attributes["named_kwargs"] = self.named_kwargs.to_dict()

        attributes["electron_outputs"] = {}
        for node_name, output in self.electron_outputs.items():
            attributes["electron_outputs"][node_name] = output.to_dict()

        return json.dumps(attributes)

    @staticmethod
    def deserialize_from_json(json_data: str) -> None:
        attributes = json.loads(json_data)

        for node_name, object_dict in attributes["electron_outputs"].items():
            attributes["electron_outputs"][node_name] = TransportableObject.from_dict(object_dict)

        attributes["named_kwargs"] = TransportableObject.from_dict(attributes["named_kwargs"])
        attributes["named_args"] = TransportableObject.from_dict(attributes["named_args"])
        attributes["inputs"] = TransportableObject.from_dict(attributes["inputs"])

        if attributes["transport_graph"]:
            tg = _TransportGraph()
            tg.deserialize_from_json(attributes["transport_graph"])
            attributes["transport_graph"] = tg

        attributes["workflow_function"] = TransportableObject.from_dict(
            attributes["workflow_function"]
        )

        def dummy_function(x):
            return x

        lat = Lattice(dummy_function)
        lat.__dict__ = attributes
        return lat

    def set_metadata(self, name: str, value: Any) -> None:
        """
        Function to add/edit metadata of given name and value
        to lattice's metadata.

        Args:
            name: Name of the metadata to be added/edited.
            value: Value of the metadata to be added/edited.

        Returns:
            None
        """

        self.metadata[name] = value

    def get_metadata(self, name: str) -> Any:
        """
        Get value of the metadata of given name.

        Args:
            name: Name of the metadata whose value is needed.

        Returns:
            value: Value of the metadata of given name.

        Raises:
            KeyError: If metadata of given name is not present.
        """

        return self.metadata.get(name, None)

    @property
    def replace_electrons(self) -> Dict[str, Callable]:
        return self.__dict__.get("_replace_electrons", {})

    @property
    def task_packing(self) -> bool:
        return self.__dict__.get("_task_packing", False)

    def build_graph(self, *args, **kwargs) -> None:
        """
        Builds the transport graph for the lattice by executing the workflow
        function which will trigger the call of all underlying electrons and
        they will get added to the transport graph for later execution.

        Also redirects any print statements inside the lattice function to null
        and ignores any exceptions caused while executing the function.

        GRAPH WILL NOT BE BUILT AFTER AN EXCEPTION HAS OCCURRED.

        Args:
            *args: Positional arguments to be passed to the workflow function.
            **kwargs: Keyword arguments to be passed to the workflow function.

        Returns:
            None

        """

        self.transport_graph.reset()

        workflow_function = self.workflow_function.get_deserialized()

        named_args, named_kwargs = get_named_params(workflow_function, args, kwargs)
        new_args = [v for _, v in named_args.items()]
        new_kwargs = dict(named_kwargs.items())

        self.inputs = TransportableObject({"args": args, "kwargs": kwargs})
        self.named_args = TransportableObject(named_args)
        self.named_kwargs = TransportableObject(named_kwargs)
        self.lattice_imports, self.cova_imports = get_imports(workflow_function)

        # Set any lattice metadata not explicitly set by the user
        constraint_names = {"executor", "workflow_executor", "hooks"}
        new_metadata = {
            name: DEFAULT_METADATA_VALUES[name]
            for name in constraint_names
            if self.metadata[name] is None
        }
        new_metadata = encode_metadata(new_metadata)

        for k, v in new_metadata.items():
            self.metadata[k] = v

        # Check whether task packing is enabled
        self._task_packing = get_config("sdk.task_packing") == "true"

        with active_lattice_manager.claim(self):
            try:
                retval = workflow_function(*new_args, **new_kwargs)
            except Exception:
                warnings.warn(
                    "Please make sure you are not manipulating an object inside the lattice."
                )
                raise

        pp = Postprocessor(lattice=self)

        with add_module_deps_to_lattice_metadata(pp, self._bound_electrons):
            if get_config("sdk.exhaustive_postprocess") == "true":
                pp.add_exhaustive_postprocess_node(self._bound_electrons.copy())
            else:
                pp.add_reconstruct_postprocess_node(retval, self._bound_electrons.copy())

        self._bound_electrons = {}  # Reset bound electrons

        # Clear this temporary attribute
        del self.__dict__["_task_packing"]

    def draw(self, *args, **kwargs) -> None:
        """
        Generate lattice graph and display in UI taking into account passed in
        arguments.

        Args:
            *args: Positional arguments to be passed to build the graph.
            **kwargs: Keyword arguments to be passed to build the graph.

        Returns:
            None
        """

        import covalent_ui.result_webhook as result_webhook

        self.build_graph(*args, **kwargs)
        result_webhook.send_draw_request(self)
        draw_preview_url = get_ui_url("/preview")
        message = f"To preview the transport graph of the lattice, visit {draw_preview_url}"
        app_log.info(message)
        print(message)
        webbrowser.open(draw_preview_url)

    def __call__(self, *args, **kwargs):
        """Execute lattice as an ordinary function for testing purposes."""

        workflow_function = self.workflow_function.get_deserialized()
        return workflow_function(*args, **kwargs)

    def dispatch(self, *args, **kwargs) -> str:
        """
        DEPRECATED: Function to dispatch workflows.

        Args:
            *args: Positional arguments for the workflow
            **kwargs: Keyword arguments for the workflow

        Returns:
            Dispatch id assigned to job
        """

        app_log.warning(
            "workflow.dispatch(your_arguments_here) is deprecated and may get removed without notice in future releases. Please use covalent.dispatch(workflow)(your_arguments_here) instead.",
            exc_info=DeprecationWarning,
        )

        from .._dispatcher_plugins import local_dispatch

        return local_dispatch(self)(*args, **kwargs)

    def dispatch_sync(self, *args, **kwargs) -> "Result":
        """
        DEPRECATED: Function to dispatch workflows synchronously by waiting for the result too.

        Args:
            *args: Positional arguments for the workflow
            **kwargs: Keyword arguments for the workflow

        Returns:
            Result of workflow execution
        """

        app_log.warning(
            "workflow.dispatch_sync(your_arguments_here) is deprecated and may get removed without notice in future releases. Please use covalent.dispatch_sync(workflow)(your_arguments_here) instead.",
            exc_info=DeprecationWarning,
        )

        from .._dispatcher_plugins import local_dispatch_sync

        return local_dispatch_sync(self)(*args, **kwargs)


def lattice(
    _func: Optional[Callable] = None,
    *,
    backend: Optional[str] = None,
    executor: Optional[Union[List[Union[str, "BaseExecutor"]], Union[str, "BaseExecutor"]]] = None,
    workflow_executor: Optional[
        Union[List[Union[str, "BaseExecutor"]], Union[str, "BaseExecutor"]]
    ] = None,
    # Add custom metadata fields here
    deps_bash: Union[DepsBash, list, str] = None,
    deps_pip: Union[DepsPip, list] = None,
    call_before: Union[List[DepsCall], DepsCall] = None,
    call_after: Union[List[DepsCall], DepsCall] = None,
    triggers: Union["BaseTrigger", List["BaseTrigger"]] = None,
    # e.g. schedule: True, whether to use a custom scheduling logic or not
) -> Lattice:
    """
    Lattice decorator to be called upon a function. Returns a new `Lattice <covalent._workflow.lattice.Lattice>` object.

    Args:
        _func: function to be decorated

    Keyword Args:
        backend: DEPRECATED: Same as `executor`.
        executor: Alternative executor object to be used in the execution of each node. If not passed, the dask
            executor is used by default.
        workflow_executor: Executor for postprocessing the workflow. Defaults to the built-in dask executor or
            the local executor depending on whether Covalent is started with the `--no-cluster` option.
        deps_bash: An optional DepsBash object specifying a list of shell commands to run before `_func`
        deps_pip: An optional DepsPip object specifying a list of PyPI packages to install before running `_func`
        call_before: An optional list of DepsCall objects specifying python functions to invoke before the electron
        call_after: An optional list of DepsCall objects specifying python functions to invoke after the electron
        triggers: Any triggers that need to be attached to this lattice, default is None

    Returns:
        :obj:`Lattice <covalent._workflow.lattice.Lattice>` : Lattice object inside which the decorated function exists.
    """

    if backend:
        app_log.warning(
            "backend is deprecated and will be removed in a future release. Please use executor keyword instead.",
            exc_info=DeprecationWarning,
        )
        executor = backend

    deps = {} if deps_bash or deps_pip else None

    if isinstance(deps_bash, DepsBash):
        deps["bash"] = deps_bash
    if isinstance(deps_bash, (list, str)):
        deps["bash"] = DepsBash(commands=deps_bash)

    if isinstance(deps_pip, DepsPip):
        deps["pip"] = deps_pip
    if isinstance(deps_pip, list):
        deps["pip"] = DepsPip(packages=deps_pip)

    if isinstance(call_before, DepsCall):
        call_before = [call_before]

    if isinstance(call_after, DepsCall):
        call_after = [call_after]

    from ..triggers import BaseTrigger

    if isinstance(triggers, BaseTrigger):
        triggers = [triggers]

    if deps is None and call_before is None and call_after is None:
        hooks = None
    else:
        hooks = {}
        if deps is not None:
            hooks["deps"] = deps
        if call_before is not None:
            hooks["call_before"] = call_before
        if call_after is not None:
            hooks["call_after"] = call_after

    constraints = {
        "executor": executor,
        "workflow_executor": workflow_executor,
        "hooks": hooks,
        "triggers": triggers,
    }

    constraints = encode_metadata(constraints)

    def decorator_lattice(func=None):
        @wraps(func)
        def wrapper_lattice(*args, **kwargs):
            lattice_object = Lattice(workflow_function=func)
            for k, v in constraints.items():
                lattice_object.set_metadata(k, v)
            lattice_object.transport_graph.lattice_metadata = lattice_object.metadata
            return lattice_object

        return wrapper_lattice()

    # Don't change the snippet below. This a subtle piece of logic that's best understood as is written.
    if _func is None:  # decorator is called with arguments
        return decorator_lattice
    else:  # decorator is called without arguments
        return decorator_lattice(_func)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy

from .deps import Deps
from .transport import TransportableObject

RESERVED_RETVAL_KEY__FILES = "files"


class DepsCall(Deps):
    """Functions, shell commands, PyPI packages, and other types of dependencies to be called in an electron's execution environment

    Deps class to encapsulate python functions to be
    called in the same execution environment as the electron.

    Attributes:
        func: A callable
        args: args list
        kwargs: kwargs dict
        retval_keyword: An optional string referencing the return value of func.

    If retval_keyword is specified, the return value of func will be
    passed during workflow execution as an argument to the electron
    corresponding to the parameter of the same name.

    Notes:
        Electron parameters to be injected during execution must have default
        parameter values.

        It is the user's responsibility to ensure that `retval_keyword` is
        actually a parameter of the electron. Unexpected behavior may occur
        otherwise.

    """

    def __init__(
        self,
        func=None,
        args=None,
        kwargs=None,
        *,
        retval_keyword="",
        override_reserved_retval_keys=False,
    ):
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        if not override_reserved_retval_keys and retval_keyword in [RESERVED_RETVAL_KEY__FILES]:
            raise RuntimeError(
                f"The retval_keyword for the specified DepsCall uses the reserved value '{retval_keyword}' please re-name to use another return value keyword."
            )

        super().__init__(
            apply_fn=func, apply_args=args, apply_kwargs=kwargs, retval_keyword=retval_keyword
        )

    def to_dict(self) -> dict:
        """Return a JSON-serializable dictionary representation of self"""
        attributes = self.__dict__.copy()
        for k, v in attributes.items():
            attributes[k] = v.to_dict() if isinstance(v, TransportableObject) else v
        return {"type": "DepsCall", "short_name": self.short_name(), "attributes": attributes}

    def from_dict(self, object_dict) -> "DepsCall":
        """Rehydrate a dictionary representation

        Args:
            object_dict: a dictionary representation returned by `to_dict`

        Returns:
            self

        Instance attributes will be overwritten.
        """

        if not object_dict:
            return self

        attributes = deepcopy(object_dict["attributes"])
        for k, v in attributes.items():
            if isinstance(v, dict) and v.get("type") == "TransportableObject":
                attributes[k] = TransportableObject.from_dict(v)
        self.__dict__ = attributes
        return self



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import site
import sys

from setuptools import Command, find_packages, setup
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info, manifest_maker

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

with open("VERSION") as f:
    version = f.read().strip()


requirements_file = "requirements.txt"
exclude_modules = [
    "tests",
    "tests.*",
]
sdk_only = os.environ.get("COVALENT_SDK_ONLY") == "1"
if sdk_only:
    requirements_file = "requirements-client.txt"
    exclude_modules += [
        "covalent_dispatcher",
        "covalent_dispatcher.*",
        "covalent_ui",
        "covalent_ui.*",
        "covalent_migrations",
    ]
    entry_points = {}
else:
    entry_points = {
        "console_scripts": [
            "covalent = covalent_dispatcher._cli.cli:cli",
        ],
    }

with open(requirements_file) as f:
    required = f.read().splitlines()

# By default we don't install qelectron requirements
# and only install them as an extra
with open("requirements-qelectron.txt") as f:
    qelectron_reqs = f.read().splitlines()


def recursively_append_files(directory: str):
    """
    Append files recursively in a given directory

    Args:
        directory: Directory within which all the subdirs and files reside

    Returns:
        filepaths: List of all file paths found recursively in the directory
    """

    filepaths = []

    for path, _, filenames in os.walk(directory):
        filepaths.extend(os.path.join(path, filename).split("/", 1)[1] for filename in filenames)

    return filepaths


def blacklist_packages():
    """Validate blacklisted packages are not installed."""

    import site
    from importlib import metadata

    if "READTHEDOCS" in os.environ:
        return

    installed_packages = [
        d.metadata["Name"] for d in metadata.distributions(path=site.getsitepackages())
    ]

    blacklist = ["cova"]

    for package in blacklist:
        if package in installed_packages:
            print("\n***************************", file=sys.stderr)
            print(
                f"Package conflict: uninstall package {package} to proceed with installation.",
                file=sys.stderr,
            )
            print("***************************\n", file=sys.stderr)
            sys.exit(1)


class BuildCovalent(build_py):
    def run(self):
        blacklist_packages()
        build_py.run(self)


class DevelopCovalent(develop):
    def run(self):
        blacklist_packages()
        develop.run(self)


class Docs(Command):
    """Generate HTML documentation"""

    description = "Generate HTML user documentation from code"

    user_options = [
        ("clean", "c", "clean directory"),
    ]

    def initialize_options(self):
        self.clean = False

    def finalize_options(self):
        pass

    def run(self):
        from doc import generate_docs

        generate_docs.run(self.clean)


class BuildUI(Command):
    """Build UI webapp"""

    description = "Build the front-end webapp from source"

    user_options = [
        ("clean", "c", "clean directory"),
    ]

    def initialize_options(self):
        self.clean = False

    def finalize_options(self):
        pass

    def run(self):
        if self.clean:
            import shutil

            shutil.rmtree("covalent_ui/webapp/build", ignore_errors=True)

        else:
            import subprocess

            subprocess.run(
                ["yarn", "install"], cwd="covalent_ui/webapp", check=True, capture_output=True
            )
            subprocess.run(
                ["yarn", "build"], cwd="covalent_ui/webapp", check=True, capture_output=True
            )


class EggInfoCovalent(egg_info):
    def find_sources(self):
        manifest_filename = os.path.join(self.egg_info, "SOURCES.txt")
        mm = manifest_maker(self.distribution)
        mm.manifest = manifest_filename

        if os.path.exists("MANIFEST.in") and sdk_only:
            with open("MANIFEST.in", "r") as f:
                lines = f.readlines()
            with open("MANIFEST_SDK.in", "w") as f:
                for line in lines:
                    if not any(excluded in line for excluded in exclude_modules):
                        f.write(line)

            mm.template = "MANIFEST_SDK.in"

        mm.run()
        self.filelist = mm.filelist

        try:
            os.remove("MANIFEST_SDK.in")
        except OSError:
            pass


setup_info = {
    "name": "covalent",
    "packages": find_packages(exclude=exclude_modules),
    "version": version,
    "maintainer": "Agnostiq",
    "url": "https://github.com/AgnostiqHQ/covalent",
    "download_url": f"https://github.com/AgnostiqHQ/covalent/archive/v{version}.tar.gz",
    "license": "Apache License 2.0",
    "author": "Agnostiq",
    "author_email": "support@agnostiq.ai",
    "description": "Covalent Workflow Tool",
    "long_description": open("README.md", encoding="utf-8").read(),
    "long_description_content_type": "text/markdown",
    "include_package_data": True,
    "zip_safe": False,
    "install_requires": required,
    "extras_require": {
        "aws": ["boto3>=1.20.48"],
        "azure": ["azure-identity>=1.13.0", "azure-storage-blob>=12.16.0"],
        "braket": ["amazon-braket-pennylane-plugin>=1.17.4", "boto3>=1.28.5"],
        "gcp": ["google-auth>=2.16.2", "google-cloud-storage>=2.7.0"],
        "mysql": ["mysqlclient>=2.1.1"],
        "postgres": ["psycopg2-binary>=2.9.5"],
        "qiskit": [
            "pennylane-qiskit==0.30",
            "qiskit==0.43.1",
            "qiskit-ibm-provider==0.6.1",
            "qiskit-ibm-runtime==0.10.0",
        ],
        "quantum": qelectron_reqs,
    },
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Adaptive Technologies",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development",
        "Topic :: System :: Distributed Computing",
    ],
    "cmdclass": {
        "egg_info": EggInfoCovalent,
        "build_py": BuildCovalent,
        "develop": DevelopCovalent,
        "docs": Docs,
        "webapp": BuildUI,
    },
    "entry_points": entry_points,
}

if __name__ == "__main__":
    setup(**setup_info)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform Electron -> ElectronSchema"""


from covalent._shared_files import logger
from covalent._shared_files.schemas.asset import AssetSchema
from covalent._shared_files.schemas.electron import (
    ElectronAssets,
    ElectronMetadata,
    ElectronSchema,
)
from covalent_dispatcher._dal.electron import ASSET_KEYS, Electron

app_log = logger.app_log


# Electrons are assumed to represent full DB records
def _export_electron_meta(e: Electron) -> ElectronMetadata:
    task_group_id = e.get_value("task_group_id", None, refresh=False)
    name = e.get_value("name", None, refresh=False)
    executor = e.get_value("executor", None, refresh=False)
    executor_data = e.get_value("executor_data", None, refresh=False)
    qelectron_data_exists = e.get_value("qelectron_data_exists", None, refresh=False)
    sub_dispatch_id = e.get_value("sub_dispatch_id", None, refresh=False)
    status = e.get_value("status", None, refresh=False)
    start_time = e.get_value("start_time", None, refresh=False)
    end_time = e.get_value("end_time", None, refresh=False)

    return ElectronMetadata(
        task_group_id=task_group_id,
        name=name,
        executor=executor,
        executor_data=executor_data,
        qelectron_data_exists=qelectron_data_exists,
        sub_dispatch_id=sub_dispatch_id,
        status=str(status),
        start_time=start_time,
        end_time=end_time,
    )


def _export_electron_assets(e: Electron) -> ElectronAssets:
    manifests = {}
    for asset_key in ASSET_KEYS:
        asset = e.assets[asset_key]
        size = asset.size
        digest_alg = asset.digest_alg
        digest = asset.digest
        scheme = asset.storage_type.value
        remote_uri = f"{scheme}://{asset.storage_path}/{asset.object_key}"
        manifests[asset_key] = AssetSchema(
            remote_uri=remote_uri, size=size, digest_alg=digest_alg, digest=digest
        )

    return ElectronAssets(**manifests)


def export_electron(e: Electron) -> ElectronSchema:
    metadata = _export_electron_meta(e)
    assets = _export_electron_assets(e)
    return ElectronSchema(id=e.node_id, metadata=metadata, assets=assets)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from pathlib import Path

from ..enums import FtCallDepReturnValue
from ..file import File


class FileTransferStrategy(ABC):
    """
    Base FileTransferStrategy class that defines the interface for file transfer strategies exposing common methods for performing copy, download, and upload operations.

    """

    # move file (from) source (to) destination
    @abstractmethod
    def cp(self, from_file: File, to_file: File) -> None:
        raise NotImplementedError

    # download here implies 'from' is a remote source
    @abstractmethod
    def download(self, from_file: File, to_file: File) -> File:
        raise NotImplementedError

    # upload here implies 'to' is a remote source
    @abstractmethod
    def upload(self, from_file: File, to_file: File) -> None:
        raise NotImplementedError

    def pre_transfer_hook(
        self,
        from_file: File,
        to_file: File,
        return_value_type: FtCallDepReturnValue = FtCallDepReturnValue.FROM_TO,
    ) -> None:
        # Create any necessary temp files needed for file transfer operations
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath
        is_from_temp_file = from_file.is_temp_file
        is_to_temp_file = to_file.is_temp_file

        # return value to be injected into kwargs of electron decorated fn as 'files'
        if return_value_type == FtCallDepReturnValue.TO:
            return_value = to_filepath
        elif return_value_type == FtCallDepReturnValue.FROM:
            return_value = from_filepath
        else:
            return_value = (from_filepath, to_filepath)

        def hook():
            if is_from_temp_file:
                from_path_obj = Path(from_filepath)
                from_path_obj.parent.mkdir(parents=True, exist_ok=True)
                from_path_obj.touch()
            if is_to_temp_file:
                to_path_obj = Path(to_filepath)
                to_path_obj.parent.mkdir(parents=True, exist_ok=True)
                to_path_obj.touch()
            return return_value

        return hook



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lattice Data Layer"""

from typing import List
from uuid import UUID

from sqlalchemy import extract, select
from sqlalchemy.orm import Session
from sqlalchemy.sql import desc, func

from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.models.dispatch_model import SortDirection
from covalent_ui.api.v1.models.lattices_model import LatticeDetailResponse


class Lattices:
    """Lattice data access layer"""

    def __init__(self, db_con: Session) -> None:
        self.db_con = db_con

    def dispatch_exist(self, dispatch_id: UUID) -> bool:
        return self.db_con.execute(
            select(Lattice).where(Lattice.dispatch_id == str(dispatch_id))
        ).fetchone()

    def get_lattices_id(self, dispatch_id: UUID) -> LatticeDetailResponse:
        """
        Get lattices from dispatch id
        Args:
            dispatch_id: Refers to the dispatch_id in lattices table
        Return:
            Top most lattice with the given dispatch_id
            (i.e lattice with the same dispatch_id, but electron_id as null)
        """

        return (
            self.db_con.query(
                Lattice.dispatch_id,
                Lattice.status,
                Lattice.storage_path.label("directory"),
                Lattice.error_filename,
                Lattice.results_filename,
                Lattice.docstring_filename,
                Lattice.started_at.label("start_time"),
                func.coalesce((Lattice.completed_at), None).label("end_time"),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
                (
                    (
                        func.coalesce(
                            extract("epoch", Lattice.completed_at),
                            extract("epoch", func.now()),
                        )
                        - extract("epoch", Lattice.started_at)
                    )
                    * 1000
                ).label("runtime"),
                func.coalesce((Lattice.updated_at), None).label("updated_at"),
            )
            .filter(Lattice.dispatch_id == str(dispatch_id), Lattice.is_active.is_not(False))
            .first()
        )

    def get_lattices_id_storage_file(self, dispatch_id: UUID):
        """
        Get storage file name
        Args:
            dispatch_id: Refers to the dispatch_id in lattices table
        Return:
            Top most lattice with the given dispatch_id along with file names
            (i.e lattice with the same dispatch_id, but electron_id as null)
        """

        return (
            self.db_con.query(
                Lattice.dispatch_id,
                Lattice.status,
                Lattice.storage_path.label("directory"),
                Lattice.error_filename,
                Lattice.function_string_filename,
                Lattice.executor,
                Lattice.executor_data,
                Lattice.workflow_executor,
                Lattice.workflow_executor_data,
                Lattice.error_filename,
                Lattice.inputs_filename,
                Lattice.results_filename,
                Lattice.storage_type,
                Lattice.function_filename,
                Lattice.started_at.label("started_at"),
                Lattice.completed_at.label("ended_at"),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
            )
            .filter(Lattice.dispatch_id == str(dispatch_id), Lattice.is_active.is_not(False))
            .first()
        )

    def get_sub_lattice_details(self, sort_by, sort_direction, dispatch_id) -> List[Lattice]:
        """
        Get summary of sub lattices
        Args:
            req.sort_by: sort by field name(run_time, status, lattice_name)
            req.direction: sort by direction ASE, DESC
        Return:
            List of sub Lattices
        """

        data = (
            self.db_con.query(
                Lattice.dispatch_id.label("dispatch_id"),
                Lattice.name.label("lattice_name"),
                (
                    (
                        func.coalesce(
                            extract("epoch", Lattice.completed_at), extract("epoch", func.now())
                        )
                        - extract("epoch", Lattice.started_at)
                    )
                    * 1000
                ).label("runtime"),
                Lattice.electron_num.label("total_electrons"),
                Lattice.completed_electron_num.label("total_electrons_completed"),
                Lattice.status.label("status"),
                Lattice.started_at.label("started_at"),
                func.coalesce((Lattice.completed_at), None).label("ended_at"),
                Lattice.updated_at.label("updated_at"),
            )
            .filter(
                Lattice.is_active.is_not(False),
                Lattice.electron_id.is_not(None),
                Lattice.root_dispatch_id == str(dispatch_id),
            )
            .order_by(
                desc(sort_by.value)
                if sort_direction == SortDirection.DESCENDING
                else sort_by.value
            )
            .all()
        )

        return data



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Triggers server API routes and standalone app definition"""


import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from inspect import signature

from fastapi import APIRouter, FastAPI, HTTPException, Request

from covalent._shared_files import logger
from covalent.triggers import BaseTrigger, available_triggers

disable_triggers = False

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router = APIRouter()
trigger_only_router = APIRouter()
triggers_only_app = FastAPI()

active_triggers = {}


def init_trigger(tr_dict: dict) -> BaseTrigger:
    """
    Recreate the trigger from its dictionary representation

    Args:
        tr_dict: Dictionary containing a representation of a Trigger

    Returns:
        Trigger object, descendant of "BaseTrigger" class
    """

    tr_name = tr_dict.pop("name")

    # Loading trigger's class
    tr_class = available_triggers[tr_name]

    # Handling required constructor params
    sig = signature(tr_class.__init__)
    init_params = {}
    for k, v in tr_dict.copy().items():
        if sig.parameters.get(k):
            init_params[k] = v
            tr_dict.pop(k)

    trigger = tr_class(**init_params)

    # Setting all other values
    for k, v in tr_dict.items():
        setattr(trigger, k, v)

    return trigger


@lru_cache
def get_threadpool():
    return ThreadPoolExecutor()


@trigger_only_router.get("/triggers/healthcheck")
async def healthcheck(request: Request):
    return {"status": "ok"}


@router.get("/triggers/status")
def trigger_server_status(request: Request):
    if disable_triggers:
        return {"status": "disabled"}
    else:
        return {"status": "running"}


@router.post("/triggers/register")
async def register_and_observe(request: Request):
    """
    Register and start the trigger's observe method
    """

    if disable_triggers:
        raise HTTPException(status_code=412, detail="Trigger endpoints are disabled as requested")

    thread_pool = get_threadpool()

    trigger_dict = await request.json()

    trigger = init_trigger(trigger_dict)

    if trigger.use_internal_funcs:
        trigger.event_loop = asyncio.get_running_loop()

    if trigger.observe_blocks:
        thread_pool.submit(trigger.observe)
    else:
        trigger.observe()

    lattice_did = trigger.lattice_dispatch_id

    if active_triggers.get(lattice_did):
        active_triggers[lattice_did].append(trigger)
    else:
        active_triggers[lattice_did] = [trigger]

    app_log.debug(f"Started trigger with id: {lattice_did}")


@router.post("/triggers/stop_observe")
async def stop_observe(request: Request):
    """
    Stop the triggers in a set of given dispatch ids
    """

    if disable_triggers:
        raise HTTPException(status_code=412, detail="Trigger endpoints are disabled as requested")

    dispatch_ids = await request.json()

    for d_id in dispatch_ids:
        for trigger in active_triggers[d_id]:
            trigger.stop()
            app_log.debug(f"Stopped observing on trigger(s) with lattice dispatch id: {d_id}")


triggers_only_app.include_router(router, prefix="/api", tags=["Triggers"])
triggers_only_app.include_router(trigger_only_router, prefix="/api", tags=["Triggers"])



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions to transform TransportGraph -> TransportGraphSchema"""

from typing import List

from covalent._shared_files import logger
from covalent._shared_files.schemas.edge import EdgeMetadata, EdgeSchema
from covalent._shared_files.schemas.electron import ElectronSchema
from covalent._shared_files.schemas.transport_graph import TransportGraphSchema
from covalent_dispatcher._dal.tg import _TransportGraph

from .electron import export_electron

app_log = logger.app_log


# Transport Graphs are assumed to be full, with a complete internal NX graph
def _export_nodes(tg: _TransportGraph) -> List[ElectronSchema]:
    g = tg.get_internal_graph_copy()
    internal_nodes = tg.get_nodes(list(g.nodes), None)
    export_nodes = [export_electron(e) for e in internal_nodes]
    return export_nodes


def _export_edges(tg: _TransportGraph) -> List[EdgeSchema]:
    edge_list = []
    g = tg.get_internal_graph_copy()
    for edge in g.edges:
        source, target, key = edge
        edge_metadata = EdgeMetadata(**g.edges[edge])
        edge_list.append(EdgeSchema(source=source, target=target, metadata=edge_metadata))

    return edge_list


def export_transport_graph(tg: _TransportGraph) -> TransportGraphSchema:
    node_list = _export_nodes(tg)
    edge_list = _export_edges(tg)
    app_log.debug(f"Exporting {len(node_list)} nodes and {len(edge_list)} edges")
    return TransportGraphSchema(nodes=node_list, links=edge_list)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Summary Routes"""

from typing import Optional

from fastapi import APIRouter, Query
from pydantic import Field
from sqlalchemy.orm import Session
from typing_extensions import Annotated

import covalent_ui.api.v1.database.config.db as db
from covalent_ui.api.v1.data_layer.summary_dal import Summary
from covalent_ui.api.v1.models.dispatch_model import (
    DeleteAllDispatchesRequest,
    DeleteDispatchesRequest,
    DeleteDispatchesResponse,
    DispatchDashBoardResponse,
    SortBy,
    SortDirection,
)
from covalent_ui.api.v1.utils.status import Status

routes: APIRouter = APIRouter()


@routes.get("/list")
def get_all_dispatches(
    count: Optional[Annotated[int, Field(gt=0, lt=100)]] = Query(10),
    offset: Optional[Annotated[int, Field(gt=-1)]] = 0,
    sort_by: Optional[SortBy] = SortBy.RUNTIME,
    search: Optional[str] = "",
    sort_direction: Optional[SortDirection] = SortDirection.DESCENDING,
    status_filter: Optional[Status] = Status.ALL,
):
    """Get All Dispatches

    Args:
        req: Dispatch Summary Request

    Returns:
        List of Dispatch Summary
    """
    with Session(db.engine) as session:
        summary = Summary(session)
        return summary.get_summary(count, offset, sort_by, search, sort_direction, status_filter)


@routes.get("/overview", response_model=DispatchDashBoardResponse)
def get_dashboard_details():
    """Overview of Dispatches

    Args:
        None

    Returns:
        An Overview of dispatches as object
    """
    with Session(db.engine) as session:
        summary = Summary(session)
        return summary.get_summary_overview()


@routes.post("/delete", response_model=DeleteDispatchesResponse)
def delete_dispatches(req: Optional[DeleteDispatchesRequest]):
    """Delete one or more Dispatches

    Args:
        req: Dispatch Dispatch Request

    Returns:
        List of deleted dispatches if success
        List of deleted dispatches if fails
    """

    with Session(db.engine) as session:
        summary = Summary(session)
        return summary.delete_dispatches(req)


@routes.post("/delete-all", response_model=DeleteDispatchesResponse)
def delete_all_dispatches(req: Optional[DeleteAllDispatchesRequest]):
    """Delete one or more Dispatches

    Args:
        req: Dispatch Dispatch Request

    Returns:
        List of deleted dispatches if success
        List of deleted dispatches if fails
    """

    with Session(db.engine) as session:
        summary = Summary(session)
        return summary.delete_all_dispatches(req)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Settings Route"""


from typing import Dict

from fastapi import APIRouter, HTTPException, status

from covalent._shared_files.config import ConfigManager
from covalent_ui.api.v1.models.settings_model import (
    GetSettingsResponseModel,
    UpdateSettingsResponseModel,
    Validators,
)

routes: APIRouter = APIRouter()


@routes.get("/settings", response_model=GetSettingsResponseModel)
def get_settings():
    """
    Read the configuration from file.
    Args:
        None
    Returns:
        Configuration file as json object.
    """
    server = {}
    client = {}

    settings = ConfigManager()

    [
        server.update({validator.value: settings.config_data[validator.value]})
        for validator in Validators
    ]
    [
        client.update({keys: settings.config_data[keys]})
        for keys in settings.config_data
        if keys not in server
    ]
    return GetSettingsResponseModel(server=server, client=client)


@routes.post("/settings", response_model=UpdateSettingsResponseModel)
def post_settings(new_entries: Dict, override_existing: bool = True):
    """
    Update the exising configuration dictionary with the configuration sent in request body.
    Only executor fields are writable.
    Args:
        new_entries: Dictionary of new entries added or updated in the config.
        override_existing: If True (default), config values from the config file
            or the input dictionary (new_entries) take precedence over any existing
            values in the config.
    Returns:
        settings updated successfully when updated.
    """
    file_type = next(iter(new_entries.keys()))
    first_key = next(iter(new_entries[file_type].keys()))
    if first_key == "":
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=[
                {
                    "loc": ["path", "update-settings"],
                    "msg": "Key should not be empty string",
                    "type": None,
                }
            ],
        )
    if len([validator.value for validator in Validators if validator.value in new_entries]) != 0:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=[
                {
                    "loc": ["path", "update-settings"],
                    "msg": "Field cannot be updated",
                    "type": None,
                }
            ],
        )
    settings = ConfigManager()
    settings.update_config(new_entries[file_type], override_existing)
    return UpdateSettingsResponseModel(data="settings updated successfully")



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Queries involving lattice
"""

from typing import Dict, List

from ..._dal.result import get_result_object
from .utils import run_in_executor


def get_sync(dispatch_id: str, keys: List[str]) -> Dict:
    refresh = False
    result_object = get_result_object(dispatch_id)
    return result_object.lattice.get_values(keys, refresh=refresh)


async def get(dispatch_id: str, keys: List[str]) -> Dict:
    return await run_in_executor(
        get_sync,
        dispatch_id,
        keys,
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel

from .asset import AssetSchema
from .common import StatusEnum
from .lattice import LATTICE_ERROR_FILENAME, LATTICE_RESULTS_FILENAME, LatticeSchema

METADATA_KEYS = {
    "start_time",
    "end_time",
    "dispatch_id",
    "root_dispatch_id",
    "status",
    "num_nodes",
}


ASSET_KEYS = {
    "result",
    "error",
}


ASSET_FILENAME_MAP = {
    "result": LATTICE_RESULTS_FILENAME,
    "error": LATTICE_ERROR_FILENAME,
}


class ResultMetadata(BaseModel):
    dispatch_id: str
    root_dispatch_id: str
    status: StatusEnum
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    # For use by redispatch
    def reset(self):
        self.dispatch_id = ""
        self.root_dispatch_id = ""
        self.status = StatusEnum.NEW_OBJECT
        self.start_time = None
        self.end_time = None


class ResultAssets(BaseModel):
    result: AssetSchema
    error: AssetSchema


class ResultSchema(BaseModel):
    metadata: ResultMetadata
    assets: ResultAssets
    lattice: LatticeSchema

    # For use by redispatch
    def reset_metadata(self):
        self.metadata.reset()

        tg = self.lattice.transport_graph
        for node in tg.nodes:
            node.metadata.reset()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..._shared_files.qelectron_utils import cloudpickle_deserialize, cloudpickle_serialize
from ..qserver import LocalQServer
from .base_client import BaseQClient

# Since in the local case, the server and client are the same
# thus the "server" class's functions are directly accessed


class LocalQClient(BaseQClient):
    def __init__(self) -> None:
        self.qserver = LocalQServer()

    @property
    def selector(self):
        return self.deserialize(self.qserver.selector)

    @selector.setter
    def selector(self, selector_func):
        self.qserver.selector = self.serialize(selector_func)

    @property
    def database(self):
        return self.deserialize(self.qserver.database)

    def submit(self, qscripts, executors, qelectron_info, qnode_specs):
        ser_qscripts = self.serialize(qscripts)
        ser_executors = self.serialize(executors)
        ser_qelectron_info = self.serialize(qelectron_info)
        ser_qnode_specs = self.serialize(qnode_specs)

        return self.qserver.submit(
            ser_qscripts, ser_executors, ser_qelectron_info, ser_qnode_specs
        )

    def get_results(self, batch_id):
        ser_results = self.qserver.get_results(batch_id)
        return self.deserialize(ser_results)

    def serialize(self, obj):
        return cloudpickle_serialize(obj)

    def deserialize(self, ser_obj):
        return cloudpickle_deserialize(ser_obj)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines the core functionality of the dispatcher
"""


from covalent._results_manager import Result

from . import runner


async def _get_task_inputs(node_id: int, node_name: str, result_object: Result) -> dict:
    """
    Return the required inputs for a task execution.
    This makes sure that any node with child nodes isn't executed twice and fetches the
    result of parent node to use as input for the child node.

    Args:
        node_id: Node id of this task in the transport graph.
        node_name: Name of the node.
        result_object: Result object to be used to update and store execution related
                       info including the results.

    Returns:
        inputs: Input dictionary to be passed to the task containing args, kwargs,
                and any parent node execution results if present.
    """

    abstract_inputs = {"args": [], "kwargs": {}}

    for parent in result_object.lattice.transport_graph.get_dependencies(node_id):
        edge_data = result_object.lattice.transport_graph.get_edge_data(parent, node_id)
        # value = result_object.lattice.transport_graph.get_node_value(parent, "output")

        for e_key, d in edge_data.items():
            if not d.get("wait_for"):
                if d["param_type"] == "arg":
                    abstract_inputs["args"].append((parent, d["arg_index"]))
                elif d["param_type"] == "kwarg":
                    key = d["edge_name"]
                    abstract_inputs["kwargs"][key] = parent

    sorted_args = sorted(abstract_inputs["args"], key=lambda x: x[1])
    abstract_inputs["args"] = [x[0] for x in sorted_args]

    input_values = await runner._get_task_input_values(result_object.dispatch_id, abstract_inputs)

    abstract_args = abstract_inputs["args"]
    abstract_kwargs = abstract_inputs["kwargs"]
    args = [input_values[node_id] for node_id in abstract_args]
    kwargs = {k: input_values[v] for k, v in abstract_kwargs.items()}
    task_input = {"args": args, "kwargs": kwargs}

    return task_input



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from furl import furl

from .._api.apiclient import CovalentAPIClient as APIClient
from .._results_manager.result import Result
from .._results_manager.results_manager import get_result, get_result_manager
from .._serialize.result import (
    extract_assets,
    merge_response_manifest,
    serialize_result,
    strip_local_uris,
)
from .._shared_files import logger
from .._shared_files.config import get_config
from .._shared_files.schemas.asset import AssetSchema
from .._shared_files.schemas.result import ResultSchema
from .._shared_files.utils import copy_file_locally, format_server_url
from .._workflow.lattice import Lattice
from ..triggers import BaseTrigger
from .base import BaseDispatcher

app_log = logger.app_log
log_stack_info = logger.log_stack_info

dispatch_cache_dir = Path(get_config("sdk.dispatch_cache_dir"))
dispatch_cache_dir.mkdir(parents=True, exist_ok=True)


def get_redispatch_request_body_v2(
    dispatch_id: str,
    staging_dir: str,
    new_args: List,
    new_kwargs: Dict,
    replace_electrons: Optional[Dict[str, Callable]],
    dispatcher_addr: str = None,
) -> ResultSchema:
    rm = get_result_manager(dispatch_id, dispatcher_addr=dispatcher_addr, wait=True)
    manifest = ResultSchema.model_validate(rm._manifest)

    # If no changes to inputs or electron, just retry the dispatch
    if not new_args and not new_kwargs and not replace_electrons:
        manifest.reset_metadata()
        app_log.debug("Resubmitting manifest only")
        return manifest

    # In all other cases we need to rebuild the graph
    rm.download_lattice_asset("workflow_function")
    rm.download_lattice_asset("workflow_function_string")
    rm.load_lattice_asset("workflow_function")
    rm.load_lattice_asset("workflow_function_string")

    if replace_electrons is None:
        replace_electrons = {}

    lat = rm.result_object.lattice

    if replace_electrons:
        lat._replace_electrons = replace_electrons

    # If lattice inputs are not supplied, retrieve them from the previous dispatch
    if not new_args and not new_kwargs:
        rm.download_lattice_asset("inputs")
        rm.load_lattice_asset("inputs")
        res_obj = rm.result_object
        inputs = res_obj.inputs.get_deserialized()
        new_args = inputs["args"]
        new_kargs = inputs["kwargs"]

    lat.build_graph(*new_args, **new_kwargs)
    if replace_electrons:
        del lat.__dict__["_replace_electrons"]

    return serialize_result(Result(lat), staging_dir)


class LocalDispatcher(BaseDispatcher):
    """
    Local dispatcher which sends the workflow to the locally running
    dispatcher server.
    """

    @staticmethod
    def dispatch(
        orig_lattice: Lattice,
        dispatcher_addr: str = None,
        *,
        disable_run: bool = False,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing
        and server address specification.

        Afterwards, send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server. If None then defaults to the address set in Covalent's config.

        Kwargs:
            disable_run: Whether to disable running the workflow and rather just save it on Covalent's server for later execution.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments
        """

        multistage = get_config("sdk.multistage_dispatch") == "true"

        # Extract triggers here
        if "triggers" in orig_lattice.metadata:
            triggers_data = orig_lattice.metadata.pop("triggers")
        else:
            triggers_data = None

        if not disable_run:
            # Determine whether to disable first run based on trigger_data
            disable_run = triggers_data is not None

        @wraps(orig_lattice)
        def wrapper(*args, **kwargs) -> str:
            """
            Send the lattice to the dispatcher server and return
            the assigned dispatch id.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The dispatch id of the workflow.
            """

            if multistage:
                dispatch_id = LocalDispatcher.register(orig_lattice, dispatcher_addr)(
                    *args, **kwargs
                )
            else:
                dispatch_id = LocalDispatcher.submit(orig_lattice, dispatcher_addr)(
                    *args, **kwargs
                )

            if triggers_data:
                LocalDispatcher.register_triggers(triggers_data, dispatch_id)

            if not disable_run:
                return LocalDispatcher.start(dispatch_id, dispatcher_addr)
            else:
                return dispatch_id

        return wrapper

    @staticmethod
    def register(
        orig_lattice: Lattice,
        dispatcher_addr: str = None,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing
        and server address specification.

        Afterwards, send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server.  If None then then defaults to the address set in Covalent's config.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        @wraps(orig_lattice)
        def wrapper(*args, **kwargs) -> str:
            """
            Send the lattice to the dispatcher server and return
            the assigned dispatch id.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The dispatch id of the workflow.
            """

            if not isinstance(orig_lattice, Lattice):
                message = f"Dispatcher expected a Lattice, received {type(orig_lattice)} instead."
                app_log.error(message)
                raise TypeError(message)

            lattice = deepcopy(orig_lattice)

            lattice.build_graph(*args, **kwargs)

            with tempfile.TemporaryDirectory() as tmp_dir:
                manifest = LocalDispatcher.prepare_manifest(lattice, tmp_dir)
                LocalDispatcher.register_manifest(manifest, dispatcher_addr)

                dispatch_id = manifest.metadata.dispatch_id

                path = dispatch_cache_dir / f"{dispatch_id}"

                with open(path, "w") as f:
                    f.write(manifest.model_dump_json())

                LocalDispatcher.upload_assets(manifest)

                return dispatch_id

        return wrapper

    @staticmethod
    def submit(
        orig_lattice: Lattice,
        dispatcher_addr: str = None,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing
        and server address specification.

        Afterwards, send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server.  If None then then defaults to the address set in Covalent's config.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        @wraps(orig_lattice)
        def wrapper(*args, **kwargs) -> str:
            """
            Send the lattice to the dispatcher server and return
            the assigned dispatch id.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The dispatch id of the workflow.
            """

            if not isinstance(orig_lattice, Lattice):
                message = f"Dispatcher expected a Lattice, received {type(orig_lattice)} instead."
                app_log.error(message)
                raise TypeError(message)

            lattice = deepcopy(orig_lattice)

            lattice.build_graph(*args, **kwargs)

            # Serialize the transport graph to JSON
            json_lattice = lattice.serialize_to_json()
            endpoint = "/api/v2/dispatches/submit"
            r = APIClient(dispatcher_addr).post(endpoint, data=json_lattice)
            r.raise_for_status()
            return r.content.decode("utf-8").strip().replace('"', "")

        return wrapper

    @staticmethod
    def start(
        dispatch_id: str,
        dispatcher_addr: str = None,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing
        and server address specification.

        Afterwards, send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server.  If None then then defaults to the address set in Covalent's config.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        endpoint = f"/api/v2/dispatches/{dispatch_id}/status"
        body = {"status": "RUNNING"}
        r = APIClient(dispatcher_addr).put(endpoint, json=body)
        r.raise_for_status()
        return r.content.decode("utf-8").strip().replace('"', "")

    @staticmethod
    def dispatch_sync(
        lattice: Lattice,
        dispatcher_addr: str = None,
    ) -> Callable:
        """
        Wrapping the synchronous dispatching functionality to allow input
        passing and server address specification.

        Afterwards, sends the lattice to the dispatcher server and return
        the result of the executed workflow.

        Args:
            orig_lattice: The lattice/workflow to send to the dispatcher server.
            dispatcher_addr: The address of the dispatcher server. If None then defaults to the address set in Covalent's config.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments.
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        @wraps(lattice)
        def wrapper(*args, **kwargs) -> Result:
            """
            Send the lattice to the dispatcher server and return
            the result of the executed workflow.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The result of the executed workflow.
            """

            return get_result(
                LocalDispatcher.dispatch(lattice, dispatcher_addr)(*args, **kwargs),
                wait=True,
            )

        return wrapper

    @staticmethod
    def redispatch(
        dispatch_id: str,
        dispatcher_addr: str = None,
        replace_electrons: Dict[str, Callable] = None,
        reuse_previous_results: bool = False,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing and server address specification.

        Args:
            dispatch_id: The dispatch id of the workflow to re-dispatch.
            dispatcher_addr: The address of the dispatcher server. If None then defaults to the address set in Covalent's config.
            replace_electrons: A dictionary of electron names and the new electron to replace them with.
            reuse_previous_results: Boolean value whether to reuse the results from the previous dispatch.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments.
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        if replace_electrons is None:
            replace_electrons = {}

        return LocalDispatcher.register_redispatch(
            dispatch_id=dispatch_id,
            dispatcher_addr=dispatcher_addr,
            replace_electrons=replace_electrons,
            reuse_previous_results=reuse_previous_results,
        )

    @staticmethod
    def register_triggers(triggers_data: List[Dict], dispatch_id: str) -> None:
        """
        Register the given triggers to the Triggers server.
        Register also starts the `observe()` method of said trigger.
        This is done by calling `BaseTrigger._register` method
        with the given trigger dictionary and is equivalent to
        calling the trigger objects `register()` method.

        Args:
            triggers_data: List of trigger dictionaries to be registered
            dispatch_id: Lattice's dispatch id to be linked with given triggers

        Returns:
            None
        """

        for tr_dict in triggers_data:
            tr_dict["lattice_dispatch_id"] = dispatch_id
            BaseTrigger._register(tr_dict)

    @staticmethod
    def stop_triggers(
        dispatch_ids: Union[str, List[str]], triggers_server_addr: str = None
    ) -> None:
        """
        Stop observing on all triggers of all given dispatch ids registered on the Triggers server.
        Args:
            dispatch_ids: Dispatch ID(s) for whose triggers are to be stopped
            triggers_server_addr: Address of the Triggers server; configured dispatcher's address is used as default
        Returns:
            None
        """

        if triggers_server_addr is None:
            triggers_server_addr = (
                "http://"
                + get_config("dispatcher.address")
                + ":"
                + str(get_config("dispatcher.port"))
            )

        if isinstance(dispatch_ids, str):
            dispatch_ids = [dispatch_ids]

        endpoint = "/api/triggers/stop_observe"
        r = APIClient(triggers_server_addr).post(endpoint, json=dispatch_ids)
        r.raise_for_status()

        app_log.debug("Triggers for following dispatch_ids have stopped observing:")
        for d_id in dispatch_ids:
            app_log.debug(d_id)

    @staticmethod
    def register_redispatch(
        dispatch_id: str,
        dispatcher_addr: str = None,
        replace_electrons: Dict[str, Callable] = None,
        reuse_previous_results: bool = False,
    ) -> Callable:
        """
        Wrapping the dispatching functionality to allow input passing and server address specification.

        Args:
            dispatch_id: The dispatch id of the workflow to re-dispatch.
            dispatcher_addr: The address of the dispatcher server. If None then defaults to the address set in Covalent's config.
            replace_electrons: A dictionary of electron names and the new electron to replace them with.
            reuse_previous_results: Boolean value whether to reuse the results from the previous dispatch.

        Returns:
            Wrapper function which takes the inputs of the workflow as arguments.
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        def func(*new_args, **new_kwargs):
            """
            Prepare the redispatch request body and redispatch the workflow.

            Args:
                *args: The inputs of the workflow.
                **kwargs: The keyword arguments of the workflow.

            Returns:
                The result of the executed workflow.
            """

            with tempfile.TemporaryDirectory() as staging_dir:
                manifest = get_redispatch_request_body_v2(
                    dispatch_id=dispatch_id,
                    staging_dir=staging_dir,
                    new_args=new_args,
                    new_kwargs=new_kwargs,
                    replace_electrons=replace_electrons,
                    dispatcher_addr=dispatcher_addr,
                )

                LocalDispatcher.register_derived_manifest(
                    manifest,
                    dispatch_id,
                    reuse_previous_results=reuse_previous_results,
                    dispatcher_addr=dispatcher_addr,
                )

                redispatch_id = manifest.metadata.dispatch_id

                path = dispatch_cache_dir / f"{redispatch_id}"

                with open(path, "w") as f:
                    f.write(manifest.model_dump_json())

                LocalDispatcher.upload_assets(manifest)

            return LocalDispatcher.start(redispatch_id, dispatcher_addr)

        return func

    @staticmethod
    def prepare_manifest(lattice, storage_path) -> ResultSchema:
        """Prepare a built-out lattice for submission"""

        result_object = Result(lattice)
        return serialize_result(result_object, storage_path)

    @staticmethod
    def register_manifest(
        manifest: ResultSchema,
        dispatcher_addr: Optional[str] = None,
        parent_dispatch_id: Optional[str] = None,
        push_assets: bool = True,
    ) -> ResultSchema:
        """Submits a manifest for registration.

        Returns:
            Dictionary representation of manifest with asset remote_uris filled in

        Side effect:
            If push_assets is False, the server will
            automatically pull the task assets from the submitted asset URIs.
        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        stripped = strip_local_uris(manifest) if push_assets else manifest
        endpoint = "/api/v2/dispatches"

        if parent_dispatch_id:
            endpoint = f"{endpoint}/{parent_dispatch_id}/subdispatches"

        r = APIClient(dispatcher_addr).post(endpoint, data=stripped.model_dump_json())
        r.raise_for_status()

        parsed_resp = ResultSchema.model_validate(r.json())

        return merge_response_manifest(manifest, parsed_resp)

    @staticmethod
    def register_derived_manifest(
        manifest: ResultSchema,
        dispatch_id: str,
        reuse_previous_results: bool = False,
        dispatcher_addr: Optional[str] = None,
    ) -> ResultSchema:
        """Submits a derived manifest for registration.

        Returns:
            Dictionary representation of manifest with asset remote_uris filled in

        """

        if dispatcher_addr is None:
            dispatcher_addr = format_server_url()

        # We don't yet support pulling assets for redispatch
        stripped = strip_local_uris(manifest)

        endpoint = f"/api/v2/dispatches/{dispatch_id}/redispatches"

        params = {"reuse_previous_results": reuse_previous_results}
        r = APIClient(dispatcher_addr).post(
            endpoint, data=stripped.model_dump_json(), params=params
        )
        r.raise_for_status()

        parsed_resp = ResultSchema.model_validate(r.json())

        return merge_response_manifest(manifest, parsed_resp)

    @staticmethod
    def upload_assets(manifest: ResultSchema):
        assets = extract_assets(manifest)
        LocalDispatcher._upload(assets)

    @staticmethod
    def _upload(assets: List[AssetSchema]):
        local_scheme_prefix = "file://"
        total = len(assets)
        number_uploaded = 0
        for i, asset in enumerate(assets):
            if not asset.remote_uri or not asset.uri:
                app_log.debug(f"Skipping asset {i+1} out of {total}")
                continue
            if asset.remote_uri.startswith(local_scheme_prefix):
                copy_file_locally(asset.uri, asset.remote_uri)
                number_uploaded += 1
            else:
                _upload_asset(asset.uri, asset.remote_uri)
                number_uploaded += 1
            app_log.debug(f"Uploaded asset {i+1} out of {total}.")
        app_log.debug(f"uploaded {number_uploaded} assets.")


def _upload_asset(local_uri, remote_uri):
    scheme_prefix = "file://"
    if local_uri.startswith(scheme_prefix):
        local_path = local_uri[len(scheme_prefix) :]
    else:
        local_path = local_uri

    with open(local_path, "rb") as reader:
        app_log.debug(f"uploading to {remote_uri}")
        f = furl(remote_uri)
        scheme = f.scheme
        host = f.host
        port = f.port
        dispatcher_addr = f"{scheme}://{host}:{port}"
        endpoint = str(f.path)
        api_client = APIClient(dispatcher_addr)

        r = api_client.put(endpoint, data=reader)
        r.raise_for_status()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Routes"""

from fastapi import APIRouter

from covalent_dispatcher._service import app, assets, runnersvc
from covalent_dispatcher._triggers_app.app import router as tr_router
from covalent_ui.api.v1.routes.end_points import (
    electron_routes,
    graph_route,
    lattice_route,
    logs_route,
    settings_routes,
    summary_routes,
)

routes = APIRouter()

dispatch_prefix = "/api/v1/dispatches"

routes.include_router(summary_routes.routes, prefix=dispatch_prefix, tags=["Dispatches"])
routes.include_router(lattice_route.routes, prefix=dispatch_prefix, tags=["Dispatches"])
routes.include_router(graph_route.routes, prefix=dispatch_prefix, tags=["Graph"])
routes.include_router(electron_routes.routes, prefix=dispatch_prefix, tags=["Electrons"])
routes.include_router(settings_routes.routes, prefix="/api/v1", tags=["Settings"])
routes.include_router(logs_route.routes, prefix="/api/v1/logs", tags=["Logs"])
routes.include_router(tr_router, prefix="/api", tags=["Triggers"])
routes.include_router(app.router, prefix="/api/v2", tags=["Dispatcher"])
routes.include_router(assets.router, prefix="/api/v2", tags=["Assets"])
routes.include_router(runnersvc.router, prefix="/api/v2", tags=["Runner"])



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# """Interface to the Jobs table"""

from typing import Any, List

from ..._db.jobdb import get_job_records, to_job_ids, update_job_records


def _set_cancel_requested(job_ids: List[int]) -> None:
    """
    Update the job record with `cancel_requested`= True (private to module)

    Arg(s)
        job_ids: List of all job ids that have been requested to be cancelled

    Return(s)
        None
    """
    records = [{"job_id": job_id, "cancel_requested": True} for job_id in job_ids]
    update_job_records(records)


async def set_cancel_requested(dispatch_id: str, task_ids: List[int]):
    """
    Set all the task with `task_ids` to be cancelled

    Arg(s)
        dispatch_id: Dispatch ID of the workflow
        task_ids: List of task ids to be cancelled

    Return(s)
        None
    """
    job_ids = to_job_ids(dispatch_id, task_ids)
    _set_cancel_requested(job_ids)


async def get_jobs_metadata(dispatch_id: str, task_ids: List[int]) -> Any:
    """
    Retrive all job records with task_ids for the given dispatch

    Arg(s)
        dispatch_id: Dispatch ID
        task_ids: List of task ids

    Return(s)
        Dictionary of job metdata associated with each task
    """
    job_ids = to_job_ids(dispatch_id, task_ids)
    return get_job_records(job_ids)


async def _set_job_metadata(dispatch_id: str, task_id: int, **kwargs) -> None:
    """
    Store any metadata associated with a job to the database

    Arg(s)
        dispatch_id: Dispatch ID of the the lattice
        task_id: ID of the task in the lattice
        kwargs: Keyword arguments

    Return(s)
        None
    """
    job_id = to_job_ids(dispatch_id, [task_id])[0]
    update_kwargs = kwargs
    update_kwargs["job_id"] = job_id
    update_job_records([update_kwargs])


async def set_job_handle(dispatch_id: str, task_id: int, job_handle: str) -> None:
    """
    Set the job handle of the task to the database

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task in the lattice
        job_handle: Job handle returned by the backend corresponding to the task (JSONable)

    Return(s)
        None
    """
    await _set_job_metadata(dispatch_id, task_id, job_handle=job_handle)


async def set_job_status(dispatch_id: str, task_id: int, status: str) -> None:
    """
    Update the status of the job in the database

    Arg(s)
        dispatch_id: Dispatch ID of the lattice
        task_id: ID of the task in the lattice
        status: status

    Return(s)
        None
    """
    await _set_job_metadata(dispatch_id, task_id, job_status=status)
